from __future__ import annotations

from collections.abc import Sequence
from typing import Annotated, Any, TypeVar, get_origin

import numpy as np
from pydantic import BaseModel, ConfigDict, Field, model_validator
from pydantic_numpy.helper.annotation import NpArrayPydanticAnnotation

from ._utils import get_dtype

TableSubclass = TypeVar("TableSubclass", bound="Table")


def _concat(value1, value2):
    if isinstance(value1, np.ndarray):
        return np.concatenate((value1, value2))
    else:
        return value1 + value2


class Table(BaseModel):
    """An abstraction of a Table of str to numpy array."""

    # You can use Table in 2 ways:
    # 1. Table(**whatever_pva_gives_us) when pvi adds a Signal to a Device that is not
    #    type hinted
    # 2. MyTable(**whatever_pva_gives_us) where the Signal is type hinted
    #
    # For 1 we want extra="allow" so it is passed through as is. There are no base class
    # fields, only "extra" fields, so they must be allowed. For 2 we want extra="forbid"
    # so it is strictly checked against the BaseModel we are supplied.
    model_config = ConfigDict(extra="allow")

    @classmethod
    def __init_subclass__(cls):
        # But forbit extra in subclasses so it gets validated
        cls.model_config = ConfigDict(validate_assignment=True, extra="forbid")
        # Change fields to have the correct annotations
        for k, anno in cls.__annotations__.items():
            if get_origin(anno) is np.ndarray:
                dtype = get_dtype(anno)
                new_anno = Annotated[
                    anno,
                    NpArrayPydanticAnnotation.factory(
                        data_type=dtype.type, dimensions=1, strict_data_typing=False
                    ),
                    Field(
                        default_factory=lambda dtype=dtype: np.array([], dtype=dtype)
                    ),
                ]
            elif get_origin(anno) is Sequence:
                new_anno = Annotated[anno, Field(default_factory=list)]
            else:
                raise TypeError(f"Cannot use annotation {anno} in a Table")
            cls.__annotations__[k] = new_anno

    def __add__(self, right: TableSubclass) -> TableSubclass:
        """Concatenate the arrays in field values."""

        if type(right) is not type(self):
            raise RuntimeError(
                f"{right} is not a `Table`, or is not the same "
                f"type of `Table` as {self}."
            )

        return type(right)(
            **{
                field_name: _concat(
                    getattr(self, field_name), getattr(right, field_name)
                )
                for field_name in self.model_fields
            }
        )

    def __eq__(self, value: object) -> bool:
        return super().__eq__(value)

    def numpy_dtype(self) -> np.dtype:
        dtype = []
        for k, v in self:
            if isinstance(v, np.ndarray):
                dtype.append((k, v.dtype))
            else:
                # TODO: use np.dtypes.StringDType when we can use in structured arrays
                # https://github.com/numpy/numpy/issues/25693
                dtype.append((k, np.dtype("S40")))
        return np.dtype(dtype)

    def numpy_table(self, selection: slice | None = None) -> np.ndarray:
        array = None
        for k, v in self:
            if selection:
                v = v[selection]
            if array is None:
                array = np.empty(v.shape, dtype=self.numpy_dtype())
            array[k] = v
        assert array is not None
        return array

    @model_validator(mode="before")
    @classmethod
    def validate_array_dtypes(cls, data: Any) -> Any:
        if isinstance(data, dict):
            data_dict = data
        elif isinstance(data, Table):
            data_dict = data.model_dump()
        else:
            raise AssertionError(f"Cannot construct Table from {data}")
        for field_name, field_value in cls.model_fields.items():
            if (
                get_origin(field_value.annotation) is np.ndarray
                and field_value.annotation
                and field_name in data_dict
            ):
                data_value = data_dict[field_name]
                expected_dtype = get_dtype(field_value.annotation)
                # Convert to correct dtype, but only if we don't lose precision
                # as a result
                cast_value = np.array(data_value).astype(expected_dtype)
                assert np.array_equal(data_value, cast_value), (
                    f"{field_name}: Cannot cast {data_value} to {expected_dtype} "
                    "without losing precision"
                )
                data_dict[field_name] = cast_value
        return data_dict

    @model_validator(mode="after")
    def validate_lengths(self) -> Table:
        lengths: dict[int, set[str]] = {}
        for field_name, field_value in self:
            lengths.setdefault(len(field_value), set()).add(field_name)
        assert len(lengths) <= 1, f"Columns should be same length, got {lengths=}"
        return self

    def __len__(self) -> int:
        return len(next(iter(self))[1])

    def __getitem__(self, item: int | slice) -> np.ndarray:
        if isinstance(item, int):
            return self.numpy_table(slice(item, item + 1))
        else:
            return self.numpy_table(item)



from typing import Annotated as A

from ophyd_async.core import (
    DeviceVector,
    SignalR,
    SignalRW,
    StandardReadable,
    StrictEnum,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import EpicsDevice, PvSuffix


class EnergyMode(StrictEnum):
    """Energy mode for `Sensor`"""

    #: Low energy mode
    low = "Low Energy"
    #: High energy mode
    high = "High Energy"


class Sensor(StandardReadable, EpicsDevice):
    """A demo sensor that produces a scalar value based on X and Y Movers"""

    value: A[SignalR[float], PvSuffix("Value"), Format.HINTED_SIGNAL]
    mode: A[SignalRW[EnergyMode], PvSuffix("Mode"), Format.CONFIG_SIGNAL]


class SensorGroup(StandardReadable):
    def __init__(self, prefix: str, name: str = "", sensor_count: int = 3) -> None:
        with self.add_children_as_readables():
            self.sensors = DeviceVector(
                {i: Sensor(f"{prefix}{i}:") for i in range(1, sensor_count + 1)}
            )

        super().__init__(name)



from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from pathlib import Path
from urllib.parse import urlunparse

from event_model import (
    ComposeStreamResource,
    ComposeStreamResourceBundle,
    StreamDatum,
    StreamRange,
    StreamResource,
)


@dataclass
class HDFDataset:
    data_key: str
    dataset: str
    shape: Sequence[int] = field(default_factory=tuple)
    dtype_numpy: str = ""
    multiplier: int = 1
    swmr: bool = False
    # Represents explicit chunk size written to disk.
    chunk_shape: tuple[int, ...] = ()


SLICE_NAME = "AD_HDF5_SWMR_SLICE"


class HDFFile:
    """
    :param full_file_name: Absolute path to the file to be written
    :param datasets: Datasets to write into the file
    """

    def __init__(
        self,
        full_file_name: Path,
        datasets: list[HDFDataset],
        hostname: str = "localhost",
    ) -> None:
        self._last_emitted = 0
        self._hostname = hostname

        if len(datasets) == 0:
            self._bundles = []
            return None

        bundler_composer = ComposeStreamResource()

        uri = urlunparse(
            (
                "file",
                self._hostname,
                str(full_file_name.absolute()),
                "",
                "",
                None,
            )
        )

        self._bundles: list[ComposeStreamResourceBundle] = [
            bundler_composer(
                mimetype="application/x-hdf5",
                uri=uri,
                data_key=ds.data_key,
                parameters={
                    "dataset": ds.dataset,
                    "swmr": ds.swmr,
                    "multiplier": ds.multiplier,
                    "chunk_shape": ds.chunk_shape,
                },
                uid=None,
                validate=True,
            )
            for ds in datasets
        ]

    def stream_resources(self) -> Iterator[StreamResource]:
        for bundle in self._bundles:
            yield bundle.stream_resource_doc

    def stream_data(self, indices_written: int) -> Iterator[StreamDatum]:
        # Indices are relative to resource
        if indices_written > self._last_emitted:
            indices: StreamRange = {
                "start": self._last_emitted,
                "stop": indices_written,
            }
            self._last_emitted = indices_written
            for bundle in self._bundles:
                yield bundle.compose_stream_datum(indices)



from collections.abc import Awaitable, Callable, Iterable
from contextlib import asynccontextmanager, contextmanager
from unittest.mock import AsyncMock, Mock

from ._device import Device, _device_mocks
from ._mock_signal_backend import MockSignalBackend
from ._signal import Signal, SignalR, _mock_signal_backends
from ._soft_signal_backend import SignalDatatypeT


def _get_mock_signal_backend(signal: Signal) -> MockSignalBackend:
    assert (
        signal in _mock_signal_backends
    ), f"Signal {signal} not connected in mock mode"
    return _mock_signal_backends[signal]


def set_mock_value(signal: Signal[SignalDatatypeT], value: SignalDatatypeT):
    """Set the value of a signal that is in mock mode."""
    backend = _get_mock_signal_backend(signal)
    backend.set_value(value)


def set_mock_put_proceeds(signal: Signal, proceeds: bool):
    """Allow or block a put with wait=True from proceeding"""
    backend = _get_mock_signal_backend(signal)

    if proceeds:
        backend.put_proceeds.set()
    else:
        backend.put_proceeds.clear()


@asynccontextmanager
async def mock_puts_blocked(*signals: Signal):
    for signal in signals:
        set_mock_put_proceeds(signal, False)
    yield
    for signal in signals:
        set_mock_put_proceeds(signal, True)


def get_mock_put(signal: Signal) -> AsyncMock:
    """Get the mock associated with the put call on the signal."""
    return _get_mock_signal_backend(signal).put_mock


def get_mock(device: Device | Signal) -> Mock:
    if isinstance(device, Signal):
        return _get_mock_signal_backend(device).mock
    return _device_mocks[device]


def reset_mock_put_calls(signal: Signal):
    backend = _get_mock_signal_backend(signal)
    backend.put_mock.reset_mock()


class _SetValuesIterator:
    # Garbage collected by the time __del__ is called unless we put it as a
    # global attrbute here.
    require_all_consumed: bool = False

    def __init__(
        self,
        signal: SignalR[SignalDatatypeT],
        values: Iterable[SignalDatatypeT],
        require_all_consumed: bool = False,
    ):
        self.signal = signal
        self.values = values
        self.require_all_consumed = require_all_consumed
        self.index = 0

        self.iterator = enumerate(values, start=1)

    def __iter__(self):
        return self

    def __next__(self):
        # Will propogate StopIteration
        self.index, next_value = next(self.iterator)
        set_mock_value(self.signal, next_value)
        return next_value

    def __del__(self):
        if self.require_all_consumed:
            # Values is cast to a list here because the user has supplied
            # require_all_consumed=True, we can therefore assume they
            # supplied a finite list.
            # In the case of require_all_consumed=False, an infinite
            # iterble is permitted
            values = list(self.values)
            if self.index != len(values):
                # Report the values consumed and the values yet to be
                # consumed
                consumed = values[0 : self.index]
                to_be_consumed = values[self.index :]
                raise AssertionError(
                    f"{self.signal.name}: {consumed} were consumed "
                    f"but {to_be_consumed} were not consumed"
                )


def set_mock_values(
    signal: SignalR[SignalDatatypeT],
    values: Iterable[SignalDatatypeT],
    require_all_consumed: bool = False,
) -> _SetValuesIterator:
    """Iterator to set a signal to a sequence of values, optionally repeating the
    sequence.

    Parameters
    ----------
    signal:
        A signal with a `MockSignalBackend` backend.
    values:
        An iterable of the values to set the signal to, on each iteration
        the value will be set.
    require_all_consumed:
        If True, an AssertionError will be raised if the iterator is deleted before
        all values have been consumed.

    Notes
    -----
    Example usage::

           for value_set in set_mock_values(signal, [1, 2, 3]):
                # do something

            cm = set_mock_values(signal, 1, 2, 3, require_all_consumed=True):
            next(cm)
            # do something
    """

    return _SetValuesIterator(
        signal,
        values,
        require_all_consumed=require_all_consumed,
    )


@contextmanager
def _unset_side_effect_cm(put_mock: AsyncMock):
    yield
    put_mock.side_effect = None


def callback_on_mock_put(
    signal: Signal[SignalDatatypeT],
    callback: Callable[[SignalDatatypeT, bool], None]
    | Callable[[SignalDatatypeT, bool], Awaitable[None]],
):
    """For setting a callback when a backend is put to.

    Can either be used in a context, with the callback being
    unset on exit, or as an ordinary function.

    Parameters
    ----------
    signal:
        A signal with a `MockSignalBackend` backend.
    callback:
        The callback to call when the backend is put to during the context.
    """
    backend = _get_mock_signal_backend(signal)
    backend.put_mock.side_effect = callback
    return _unset_side_effect_cm(backend.put_mock)



from ._signal import (
    infer_python_type,
    infer_signal_type,
    make_backend,
    tango_signal_r,
    tango_signal_rw,
    tango_signal_w,
    tango_signal_x,
)
from ._tango_transport import (
    AttributeProxy,
    CommandProxy,
    TangoSignalBackend,
    ensure_proper_executor,
    get_dtype_extended,
    get_python_type,
    get_tango_trl,
    get_trl_descriptor,
)

__all__ = (
    "AttributeProxy",
    "CommandProxy",
    "ensure_proper_executor",
    "TangoSignalBackend",
    "get_python_type",
    "get_dtype_extended",
    "get_trl_descriptor",
    "get_tango_trl",
    "infer_python_type",
    "infer_signal_type",
    "make_backend",
    "tango_signal_r",
    "tango_signal_rw",
    "tango_signal_w",
    "tango_signal_x",
)



import asyncio
from enum import Enum

import numpy as np
import numpy.typing as npt
import pytest
from test_base_device import TestDevice
from test_tango_signals import (
    EchoDevice,
    make_backend,
    prepare_device,
)

from ophyd_async.core import (
    NotConnected,
)
from ophyd_async.tango import (
    AttributeProxy,
    CommandProxy,
    TangoSignalBackend,
    ensure_proper_executor,
    get_dtype_extended,
    get_python_type,
    get_tango_trl,
    get_trl_descriptor,
)
from tango import (
    CmdArgType,
    DevState,
)
from tango.asyncio import DeviceProxy
from tango.asyncio_executor import (
    AsyncioExecutor,
    get_global_executor,
    set_global_executor,
)
from tango.test_context import MultiDeviceTestContext


# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def tango_test_device():
    with MultiDeviceTestContext(
        [{"class": TestDevice, "devices": [{"name": "test/device/1"}]}], process=True
    ) as context:
        yield context.get_device_access("test/device/1")


# --------------------------------------------------------------------
@pytest.fixture(scope="module")
def echo_device():
    with MultiDeviceTestContext(
        [{"class": EchoDevice, "devices": [{"name": "test/device/1"}]}], process=True
    ) as context:
        yield context.get_device_access("test/device/1")


# --------------------------------------------------------------------
@pytest.fixture(autouse=True)
def reset_tango_asyncio():
    set_global_executor(None)


# --------------------------------------------------------------------
class HelperClass:
    @ensure_proper_executor
    async def mock_func(self):
        return "executed"


# Test function
@pytest.mark.asyncio
async def test_ensure_proper_executor():
    # Instantiate the helper class and call the decorated method
    helper_instance = HelperClass()
    result = await helper_instance.mock_func()

    # Assertions
    assert result == "executed"
    assert isinstance(get_global_executor(), AsyncioExecutor)


# --------------------------------------------------------------------
@pytest.mark.parametrize(
    "tango_type, expected",
    [
        (CmdArgType.DevVoid, (False, None, "string")),
        (CmdArgType.DevBoolean, (False, bool, "integer")),
        (CmdArgType.DevShort, (False, int, "integer")),
        (CmdArgType.DevLong, (False, int, "integer")),
        (CmdArgType.DevFloat, (False, float, "number")),
        (CmdArgType.DevDouble, (False, float, "number")),
        (CmdArgType.DevUShort, (False, int, "integer")),
        (CmdArgType.DevULong, (False, int, "integer")),
        (CmdArgType.DevString, (False, str, "string")),
        (CmdArgType.DevVarCharArray, (True, list[str], "string")),
        (CmdArgType.DevVarShortArray, (True, int, "integer")),
        (CmdArgType.DevVarLongArray, (True, int, "integer")),
        (CmdArgType.DevVarFloatArray, (True, float, "number")),
        (CmdArgType.DevVarDoubleArray, (True, float, "number")),
        (CmdArgType.DevVarUShortArray, (True, int, "integer")),
        (CmdArgType.DevVarULongArray, (True, int, "integer")),
        (CmdArgType.DevVarStringArray, (True, str, "string")),
        # (CmdArgType.DevVarLongStringArray, (True, str, "string")),
        # (CmdArgType.DevVarDoubleStringArray, (True, str, "string")),
        (CmdArgType.DevState, (False, CmdArgType.DevState, "string")),
        (CmdArgType.ConstDevString, (False, str, "string")),
        (CmdArgType.DevVarBooleanArray, (True, bool, "integer")),
        (CmdArgType.DevUChar, (False, int, "integer")),
        (CmdArgType.DevLong64, (False, int, "integer")),
        (CmdArgType.DevULong64, (False, int, "integer")),
        (CmdArgType.DevVarLong64Array, (True, int, "integer")),
        (CmdArgType.DevVarULong64Array, (True, int, "integer")),
        (CmdArgType.DevEncoded, (False, list[str], "string")),
        (CmdArgType.DevEnum, (False, Enum, "string")),
        # (CmdArgType.DevPipeBlob, (False, list[str], "string")),
        (float, (False, float, "number")),
    ],
)
def test_get_python_type(tango_type, expected):
    if tango_type is not float:
        assert get_python_type(tango_type) == expected
    else:
        # get_python_type should raise a TypeError
        with pytest.raises(TypeError) as exc_info:
            get_python_type(tango_type)
        assert str(exc_info.value) == "Unknown TangoType"


# --------------------------------------------------------------------
@pytest.mark.parametrize(
    "datatype, expected",
    [
        (npt.NDArray[np.float64], np.dtype("float64")),
        (npt.NDArray[np.int8], np.dtype("int8")),
        (npt.NDArray[np.uint8], np.dtype("uint8")),
        (npt.NDArray[np.int32], np.dtype("int32")),
        (npt.NDArray[np.int64], np.dtype("int64")),
        (npt.NDArray[np.uint16], np.dtype("uint16")),
        (npt.NDArray[np.uint32], np.dtype("uint32")),
        (npt.NDArray[np.uint64], np.dtype("uint64")),
        (npt.NDArray[np.bool_], np.dtype("bool")),
        (npt.NDArray[DevState], CmdArgType.DevState),
        (npt.NDArray[np.str_], np.dtype("str")),
        (npt.NDArray[np.float32], np.dtype("float32")),
        (npt.NDArray[np.complex64], np.dtype("complex64")),
        (npt.NDArray[np.complex128], np.dtype("complex128")),
    ],
)
def test_get_dtype_extended(datatype, expected):
    assert get_dtype_extended(datatype) == expected


# --------------------------------------------------------------------


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "datatype, tango_resource, expected_descriptor",
    [
        (
            int,
            "test/device/1/justvalue",
            {"source": "test/device/1/justvalue", "dtype": "integer", "shape": []},
        ),
        (
            float,
            "test/device/1/limitedvalue",
            {"source": "test/device/1/limitedvalue", "dtype": "number", "shape": []},
        ),
        (
            npt.NDArray[float],
            "test/device/1/array",
            {"source": "test/device/1/array", "dtype": "array", "shape": [2, 3]},
        ),
        # Add more test cases as needed
    ],
)
async def test_get_trl_descriptor(
    tango_test_device, datatype, tango_resource, expected_descriptor
):
    proxy = await DeviceProxy(tango_test_device)
    tr_configs = {
        tango_resource.split("/")[-1]: await proxy.get_attribute_config(
            tango_resource.split("/")[-1]
        )
    }
    descriptor = get_trl_descriptor(datatype, tango_resource, tr_configs)
    assert descriptor == expected_descriptor


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "trl, proxy_needed, expected_type, should_raise",
    [
        ("test/device/1/justvalue", True, AttributeProxy, False),
        ("test/device/1/justvalue", False, AttributeProxy, False),
        ("test/device/1/clear", True, CommandProxy, False),
        ("test/device/1/clear", False, CommandProxy, False),
        ("test/device/1/nonexistent", True, None, True),
    ],
)
async def test_get_tango_trl(
    tango_test_device, trl, proxy_needed, expected_type, should_raise
):
    proxy = await DeviceProxy(tango_test_device) if proxy_needed else None
    if should_raise:
        with pytest.raises(RuntimeError):
            await get_tango_trl(trl, proxy, 1)
    else:
        result = await get_tango_trl(trl, proxy, 1)
        assert isinstance(result, expected_type)


# --------------------------------------------------------------------


@pytest.mark.asyncio
@pytest.mark.parametrize("attr", ["justvalue", "array"])
async def test_attribute_proxy_get(tango_test_device, attr):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)
    val = None
    val = await attr_proxy.get()
    assert val is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "attr, wait",
    [("justvalue", True), ("justvalue", False), ("array", True), ("array", False)],
)
async def test_attribute_proxy_put(tango_test_device, attr, wait):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)

    old_value = await attr_proxy.get()
    new_value = old_value + 1
    status = await attr_proxy.put(new_value, wait=wait, timeout=0.1)
    if status:
        await status
    else:
        if not wait:
            raise AssertionError("If wait is False, put should return a status object")
    await asyncio.sleep(1.0)
    updated_value = await attr_proxy.get()
    if isinstance(new_value, np.ndarray):
        assert np.all(updated_value == new_value)
    else:
        assert updated_value == new_value


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("wait", [True, False])
async def test_attribute_proxy_put_force_timeout(tango_test_device, wait):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "slow_attribute")
    with pytest.raises(TimeoutError) as exc_info:
        status = await attr_proxy.put(3.0, wait=wait, timeout=0.1)
        await status
    assert "attr put failed" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("wait", [True, False])
async def test_attribute_proxy_put_exceptions(tango_test_device, wait):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "raise_exception_attr")
    with pytest.raises(RuntimeError) as exc_info:
        status = await attr_proxy.put(3.0, wait=wait)
        await status
    assert "device failure" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize(
    "attr, new_value", [("justvalue", 10), ("array", np.array([[2, 3, 4], [5, 6, 7]]))]
)
async def test_attribute_proxy_get_w_value(tango_test_device, attr, new_value):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)
    await attr_proxy.put(new_value)
    await asyncio.sleep(1.0)
    attr_proxy_value = await attr_proxy.get()
    if isinstance(new_value, np.ndarray):
        assert np.all(attr_proxy_value == new_value)
    else:
        assert attr_proxy_value == new_value


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_get_config(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    config = await attr_proxy.get_config()
    assert config.writable is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_get_reading(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    reading = await attr_proxy.get_reading()
    assert reading["value"] is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_has_subscription(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    expected = bool(attr_proxy._callback)
    has_subscription = attr_proxy.has_subscription()
    assert has_subscription is expected


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_subscribe_callback(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    backend = await make_backend(float, source)
    attr_proxy = backend.proxies[source]
    val = None

    def callback(reading):
        nonlocal val
        val = reading["value"]

    attr_proxy.subscribe_callback(callback)
    assert attr_proxy.has_subscription()
    old_value = await attr_proxy.get()
    new_value = old_value + 1
    await attr_proxy.put(new_value)
    await asyncio.sleep(0.2)
    attr_proxy.unsubscribe_callback()
    assert val == new_value

    attr_proxy.set_polling(False)
    attr_proxy.support_events = False
    with pytest.raises(RuntimeError) as exc_info:
        attr_proxy.subscribe_callback(callback)
    assert "Cannot set a callback" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_unsubscribe_callback(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    backend = await make_backend(float, source)
    attr_proxy = backend.proxies[source]

    def callback(reading):
        pass

    attr_proxy.subscribe_callback(callback)
    assert attr_proxy.has_subscription()
    attr_proxy.unsubscribe_callback()
    assert not attr_proxy.has_subscription()


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_set_polling(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "justvalue")
    attr_proxy.set_polling(True, 0.1, 1, 0.1)
    assert attr_proxy._allow_polling
    assert attr_proxy._polling_period == 0.1
    assert attr_proxy._abs_change == 1
    assert attr_proxy._rel_change == 0.1
    attr_proxy.set_polling(False)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_poll(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, "floatvalue")
    attr_proxy.support_events = False

    def callback(reading):
        nonlocal val
        val = reading["value"]

    def bad_callback():
        pass

    # Test polling with absolute change
    val = None

    attr_proxy.set_polling(True, 0.1, 1, 1.0)
    attr_proxy.subscribe_callback(callback)
    current_value = await attr_proxy.get()
    new_value = current_value + 2
    await attr_proxy.put(new_value)
    polling_period = attr_proxy._polling_period
    await asyncio.sleep(polling_period)
    assert val is not None
    attr_proxy.unsubscribe_callback()

    # Test polling with relative change
    val = None
    attr_proxy.set_polling(True, 0.1, 100, 0.1)
    attr_proxy.subscribe_callback(callback)
    current_value = await attr_proxy.get()
    new_value = current_value * 2
    await attr_proxy.put(new_value)
    polling_period = attr_proxy._polling_period
    await asyncio.sleep(polling_period)
    assert val is not None
    attr_proxy.unsubscribe_callback()

    # Test polling with small changes. This should not update last_reading
    attr_proxy.set_polling(True, 0.1, 100, 1.0)
    attr_proxy.subscribe_callback(callback)
    await asyncio.sleep(0.2)
    current_value = await attr_proxy.get()
    new_value = current_value + 1
    val = None
    await attr_proxy.put(new_value)
    polling_period = attr_proxy._polling_period
    await asyncio.sleep(polling_period * 2)
    assert val is None
    attr_proxy.unsubscribe_callback()

    # Test polling with bad callback
    attr_proxy.subscribe_callback(bad_callback)
    await asyncio.sleep(0.2)
    assert "Could not poll the attribute" in str(attr_proxy.exception)
    attr_proxy.unsubscribe_callback()


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("attr", ["array", "label"])
async def test_attribute_poll_stringsandarrays(tango_test_device, attr):
    device_proxy = await DeviceProxy(tango_test_device)
    attr_proxy = AttributeProxy(device_proxy, attr)
    attr_proxy.support_events = False

    def callback(reading):
        nonlocal val
        val = reading["value"]

    val = None
    attr_proxy.set_polling(True, 0.1)
    attr_proxy.subscribe_callback(callback)
    await asyncio.sleep(0.2)
    assert val is not None
    if isinstance(val, np.ndarray):
        await attr_proxy.put(np.array([[2, 3, 4], [5, 6, 7]]))
        await asyncio.sleep(0.5)
        assert np.all(val == np.array([[2, 3, 4], [5, 6, 7]]))
    if isinstance(val, str):
        await attr_proxy.put("new label")
        await asyncio.sleep(0.5)
        assert val == "new label"


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_attribute_poll_exceptions(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    # Try to poll a non-existent attribute
    attr_proxy = AttributeProxy(device_proxy, "nonexistent")
    attr_proxy.support_events = False
    attr_proxy.set_polling(True, 0.1)

    def callback(reading, value):
        pass

    attr_proxy.subscribe_callback(callback)
    await asyncio.sleep(0.2)
    assert "Could not poll the attribute" in str(attr_proxy.exception)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_proxy_put_wait(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "clear")

    cmd_proxy._last_reading = None
    await cmd_proxy.put(None, wait=True)
    assert cmd_proxy._last_reading["value"] == "Received clear command"

    # Force timeout
    cmd_proxy = CommandProxy(device_proxy, "slow_command")
    cmd_proxy._last_reading = None
    with pytest.raises(TimeoutError) as exc_info:
        await cmd_proxy.put(None, wait=True, timeout=0.1)
    assert "command failed" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_proxy_put_nowait(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "slow_command")

    # Reply before timeout
    cmd_proxy._last_reading = None
    status = await cmd_proxy.put(None, wait=False, timeout=0.5)
    assert cmd_proxy._last_reading is None
    await status
    assert cmd_proxy._last_reading["value"] == "Completed slow command"

    # Timeout
    cmd_proxy._last_reading = None
    status = await cmd_proxy.put(None, wait=False, timeout=0.1)
    with pytest.raises(TimeoutError) as exc_info:
        await status
    assert str(exc_info.value) == "Timeout while waiting for command reply"

    # No timeout
    cmd_proxy._last_reading = None
    status = await cmd_proxy.put(None, wait=False)
    assert cmd_proxy._last_reading is None
    await status
    assert cmd_proxy._last_reading["value"] == "Completed slow command"


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("wait", [True, False])
async def test_command_proxy_put_exceptions(tango_test_device, wait):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "raise_exception_cmd")
    with pytest.raises(RuntimeError) as exc_info:
        await cmd_proxy.put(None, wait=True)
    assert "device failure" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_get(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "clear")
    await cmd_proxy.put(None, wait=True, timeout=1.0)
    reading = cmd_proxy._last_reading
    assert reading["value"] is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_get_config(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "clear")
    config = await cmd_proxy.get_config()
    assert config.out_type is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_get_reading(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "clear")
    await cmd_proxy.put(None, wait=True, timeout=1.0)
    reading = await cmd_proxy.get_reading()
    assert reading["value"] is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_command_set_polling(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    cmd_proxy = CommandProxy(device_proxy, "clear")
    cmd_proxy.set_polling(True, 0.1)
    # Set polling in the command proxy currently does nothing
    assert True


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_init(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)
    assert transport is not None


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_source(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source)
    transport_source = transport.source("", True)
    assert transport_source == source


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_datatype_allowed(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    backend = await make_backend(float, source)

    assert backend.datatype_allowed(int)
    assert backend.datatype_allowed(float)
    assert backend.datatype_allowed(str)
    assert backend.datatype_allowed(bool)
    assert backend.datatype_allowed(np.ndarray)
    assert backend.datatype_allowed(Enum)
    assert backend.datatype_allowed(DevState)
    assert not backend.datatype_allowed(list)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_connect(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    backend = await make_backend(float, source, connect=False)
    assert backend is not None
    await backend.connect(1)
    backend.read_trl = ""
    with pytest.raises(RuntimeError) as exc_info:
        await backend.connect(1)
    assert "trl not set" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_connect_and_store_config(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)
    await transport._connect_and_store_config(source, 1)
    assert transport.trl_configs[source] is not None

    with pytest.raises(RuntimeError) as exc_info:
        await transport._connect_and_store_config("", 1)
    assert "trl not set" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_put(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnected) as exc_info:
        await transport.put(1.0)
    assert "Not connected" in str(exc_info.value)

    await transport.connect(1)
    source = transport.source("", True)
    await transport.put(2.0)
    val = await transport.proxies[source].get_w_value()
    assert val == 2.0


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_get_datakey(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)
    await transport.connect(1)
    datakey = await transport.get_datakey(source)
    assert datakey["source"] == source
    assert datakey["dtype"] == "number"
    assert datakey["shape"] == []


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_get_reading(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnected) as exc_info:
        await transport.put(1.0)
    assert "Not connected" in str(exc_info.value)

    await transport.connect(1)
    reading = await transport.get_reading()
    assert reading["value"] == 1.0


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_get_value(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnected) as exc_info:
        await transport.put(1.0)
    assert "Not connected" in str(exc_info.value)

    await transport.connect(1)
    value = await transport.get_value()
    assert value == 1.0


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_get_setpoint(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnected) as exc_info:
        await transport.put(1.0)
    assert "Not connected" in str(exc_info.value)

    await transport.connect(1)
    new_setpoint = 2.0
    await transport.put(new_setpoint)
    setpoint = await transport.get_setpoint()
    assert setpoint == new_setpoint


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_set_callback(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)

    with pytest.raises(NotConnected) as exc_info:
        await transport.put(1.0)
    assert "Not connected" in str(exc_info.value)

    await transport.connect(1)
    val = None

    def callback(reading):
        nonlocal val
        val = reading["value"]

    # Correct usage
    transport.set_callback(callback)
    current_value = await transport.get_value()
    new_value = current_value + 2
    await transport.put(new_value)
    await asyncio.sleep(0.1)
    assert val == new_value

    # Try to add second callback
    with pytest.raises(RuntimeError) as exc_info:
        transport.set_callback(callback)
    assert "Cannot set a callback when one is already set"

    transport.set_callback(None)

    # Try to add a callback to a non-callable proxy
    transport.allow_events(False)
    transport.set_polling(False)
    with pytest.raises(RuntimeError) as exc_info:
        transport.set_callback(callback)
    assert "Cannot set event" in str(exc_info.value)

    # Try to add a non-callable callback
    transport.allow_events(True)
    with pytest.raises(RuntimeError) as exc_info:
        transport.set_callback(1)
    assert "Callback must be a callable" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_set_polling(echo_device):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)
    transport.set_polling(True, 0.1, 1, 0.1)
    assert transport._polling == (True, 0.1, 1, 0.1)


# --------------------------------------------------------------------
@pytest.mark.asyncio
@pytest.mark.parametrize("allow", [True, False])
async def test_tango_transport_allow_events(echo_device, allow):
    await prepare_device(echo_device, "float_scalar_attr", 1.0)
    source = echo_device + "/" + "float_scalar_attr"
    transport = await make_backend(float, source, connect=False)
    transport.allow_events(allow)
    assert transport.support_events == allow


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_read_and_write_trl(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    # Must use a FQTRL, at least on windows.
    read_trl = tango_test_device + "/" + "readback"
    write_trl = tango_test_device + "/" + "setpoint"

    # Test with existing proxy
    transport = TangoSignalBackend(float, read_trl, write_trl, device_proxy)
    await transport.connect(1)
    reading = await transport.get_reading()
    initial_value = reading["value"]
    new_value = initial_value + 1.0
    await transport.put(new_value)
    updated_value = await transport.get_value()
    assert updated_value == new_value

    # Without pre-existing proxy
    transport = TangoSignalBackend(float, read_trl, write_trl, None)
    await transport.connect(1)
    reading = await transport.get_reading()
    initial_value = reading["value"]
    new_value = initial_value + 1.0
    await transport.put(new_value)
    updated_value = await transport.get_value()
    assert updated_value == new_value


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_read_only_trl(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = device_proxy.dev_name()
    read_trl = trl + "/" + "readonly"

    # Test with existing proxy
    transport = TangoSignalBackend(int, read_trl, read_trl, device_proxy)
    await transport.connect(1)
    with pytest.raises(RuntimeError) as exc_info:
        await transport.put(1)
    assert "is not writable" in str(exc_info.value)


# --------------------------------------------------------------------
@pytest.mark.asyncio
async def test_tango_transport_nonexistent_trl(tango_test_device):
    device_proxy = await DeviceProxy(tango_test_device)
    trl = device_proxy.dev_name()
    nonexistent_trl = trl + "/" + "nonexistent"

    # Test with existing proxy
    transport = TangoSignalBackend(int, nonexistent_trl, nonexistent_trl, device_proxy)
    with pytest.raises(RuntimeError) as exc_info:
        await transport.connect(1)
    assert "cannot be found" in str(exc_info.value)

    # Without pre-existing proxy
    transport = TangoSignalBackend(int, nonexistent_trl, nonexistent_trl, None)
    with pytest.raises(RuntimeError) as exc_info:
        await transport.connect(1)
    assert "cannot be found" in str(exc_info.value)



"""Tango Signals over Pytango"""

from __future__ import annotations

import logging
from enum import Enum, IntEnum

import numpy.typing as npt

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Signal,
    SignalDatatypeT,
    SignalR,
    SignalRW,
    SignalW,
    SignalX,
)
from tango import AttrDataFormat, AttrWriteType, CmdArgType, DeviceProxy, DevState
from tango.asyncio import DeviceProxy as AsyncDeviceProxy

from ._tango_transport import TangoSignalBackend, get_python_type


def make_backend(
    datatype: type[SignalDatatypeT] | None,
    read_trl: str = "",
    write_trl: str = "",
    device_proxy: DeviceProxy | None = None,
) -> TangoSignalBackend:
    return TangoSignalBackend(datatype, read_trl, write_trl, device_proxy)


def tango_signal_rw(
    datatype: type[SignalDatatypeT],
    read_trl: str,
    write_trl: str = "",
    device_proxy: DeviceProxy | None = None,
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalRW[SignalDatatypeT]:
    """Create a `SignalRW` backed by 1 or 2 Tango Attribute/Command

    Parameters
    ----------
    datatype:
        Check that the Attribute/Command is of this type
    read_trl:
        The Attribute/Command to read and monitor
    write_trl:
        If given, use this Attribute/Command to write to, otherwise use read_trl
    device_proxy:
        If given, this DeviceProxy will be used
    timeout:
        The timeout for the read and write operations
    name:
        The name of the Signal
    """
    backend = make_backend(datatype, read_trl, write_trl or read_trl, device_proxy)
    return SignalRW(backend, timeout=timeout, name=name)


def tango_signal_r(
    datatype: type[SignalDatatypeT],
    read_trl: str,
    device_proxy: DeviceProxy | None = None,
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalR[SignalDatatypeT]:
    """Create a `SignalR` backed by 1 Tango Attribute/Command

    Parameters
    ----------
    datatype:
        Check that the Attribute/Command is of this type
    read_trl:
        The Attribute/Command to read and monitor
    device_proxy:
        If given, this DeviceProxy will be used
    timeout:
        The timeout for the read operation
    name:
        The name of the Signal
    """
    backend = make_backend(datatype, read_trl, read_trl, device_proxy)
    return SignalR(backend, timeout=timeout, name=name)


def tango_signal_w(
    datatype: type[SignalDatatypeT],
    write_trl: str,
    device_proxy: DeviceProxy | None = None,
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalW[SignalDatatypeT]:
    """Create a `SignalW` backed by 1 Tango Attribute/Command

    Parameters
    ----------
    datatype:
        Check that the Attribute/Command is of this type
    write_trl:
        The Attribute/Command to write to
    device_proxy:
        If given, this DeviceProxy will be used
    timeout:
        The timeout for the write operation
    name:
        The name of the Signal
    """
    backend = make_backend(datatype, write_trl, write_trl, device_proxy)
    return SignalW(backend, timeout=timeout, name=name)


def tango_signal_x(
    write_trl: str,
    device_proxy: DeviceProxy | None = None,
    timeout: float = DEFAULT_TIMEOUT,
    name: str = "",
) -> SignalX:
    """Create a `SignalX` backed by 1 Tango Attribute/Command

    Parameters
    ----------
    write_trl:
        The Attribute/Command to write its initial value to on execute
    device_proxy:
        If given, this DeviceProxy will be used
    timeout:
        The timeout for the command operation
    name:
        The name of the Signal
    """
    backend = make_backend(None, write_trl, write_trl, device_proxy)
    return SignalX(backend, timeout=timeout, name=name)


async def infer_python_type(
    trl: str = "", proxy: DeviceProxy | None = None
) -> object | npt.NDArray | type[DevState] | IntEnum:
    # TODO: work out if this is still needed
    device_trl, tr_name = trl.rsplit("/", 1)
    if proxy is None:
        dev_proxy = await AsyncDeviceProxy(device_trl)
    else:
        dev_proxy = proxy

    if tr_name in dev_proxy.get_command_list():
        config = await dev_proxy.get_command_config(tr_name)
        isarray, py_type, _ = get_python_type(config.in_type)
    elif tr_name in dev_proxy.get_attribute_list():
        config = await dev_proxy.get_attribute_config(tr_name)
        isarray, py_type, _ = get_python_type(config.data_type)
        if py_type is Enum:
            enum_dict = {label: i for i, label in enumerate(config.enum_labels)}
            py_type = IntEnum("TangoEnum", enum_dict)
        if config.data_format in [AttrDataFormat.SPECTRUM, AttrDataFormat.IMAGE]:
            isarray = True
    else:
        raise RuntimeError(f"Cannot find {tr_name} in {device_trl}")

    if py_type is CmdArgType.DevState:
        py_type = DevState

    return npt.NDArray[py_type] if isarray else py_type


async def infer_signal_type(
    trl, proxy: DeviceProxy | None = None
) -> type[Signal] | None:
    device_trl, tr_name = trl.rsplit("/", 1)
    if proxy is None:
        dev_proxy = await AsyncDeviceProxy(device_trl)
    else:
        dev_proxy = proxy

    if tr_name in dev_proxy.get_pipe_list():
        raise NotImplementedError("Pipes are not supported")

    if tr_name not in dev_proxy.get_attribute_list():
        if tr_name not in dev_proxy.get_command_list():
            raise RuntimeError(f"Cannot find {tr_name} in {device_trl}")

    if tr_name in dev_proxy.get_attribute_list():
        config = await dev_proxy.get_attribute_config(tr_name)
        if config.writable in [AttrWriteType.READ_WRITE, AttrWriteType.READ_WITH_WRITE]:
            return SignalRW
        elif config.writable == AttrWriteType.READ:
            return SignalR
        else:
            return SignalW

    if tr_name in dev_proxy.get_command_list():
        config = await dev_proxy.get_command_config(tr_name)
        if config.in_type == CmdArgType.DevVoid:
            return SignalX
        elif config.in_type != config.out_type:
            logging.debug("Commands with different in and out dtypes are not supported")
            return None
        else:
            return SignalRW
    raise RuntimeError(f"Unable to infer signal character for {trl}")



from unittest.mock import ANY, AsyncMock, MagicMock

import pytest

from ophyd_async.core import DetectorTrigger, DeviceCollector, get_mock_put
from ophyd_async.epics.eiger import EigerDetector, EigerTriggerInfo


@pytest.fixture
def detector(RE):
    with DeviceCollector(mock=True):
        detector = EigerDetector("BL03I", MagicMock())
    return detector


def test_when_detector_initialised_then_driver_and_odin_have_expected_prefixes(
    detector,
):
    assert "BL03I-EA-EIGER-01:" in detector.drv.arm.source
    assert "BL03I-EA-ODIN-01:FP:" in detector.odin.acquisition_id.source


async def test_when_prepared_with_energy_then_energy_set_on_detector(detector):
    detector.controller.arm = AsyncMock()
    await detector.prepare(
        EigerTriggerInfo(
            frame_timeout=None,
            number_of_triggers=1,
            trigger=DetectorTrigger.internal,
            deadtime=None,
            livetime=None,
            energy_ev=10000,
        )
    )

    get_mock_put(detector.drv.photon_energy).assert_called_once_with(10000, wait=ANY)



from ._pilatus import PilatusDetector, PilatusReadoutTime
from ._pilatus_controller import PilatusController
from ._pilatus_io import PilatusDriverIO, PilatusTriggerMode

__all__ = [
    "PilatusDetector",
    "PilatusReadoutTime",
    "PilatusController",
    "PilatusDriverIO",
    "PilatusTriggerMode",
]



import asyncio

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DetectorController,
    DetectorTrigger,
    TriggerInfo,
    wait_for_value,
)
from ophyd_async.epics import adcore

from ._pilatus_io import PilatusDriverIO, PilatusTriggerMode


class PilatusController(DetectorController):
    _supported_trigger_types = {
        DetectorTrigger.internal: PilatusTriggerMode.internal,
        DetectorTrigger.constant_gate: PilatusTriggerMode.ext_enable,
        DetectorTrigger.variable_gate: PilatusTriggerMode.ext_enable,
    }

    def __init__(
        self,
        driver: PilatusDriverIO,
        readout_time: float,
    ) -> None:
        self._drv = driver
        self._readout_time = readout_time
        self._arm_status: AsyncStatus | None = None

    def get_deadtime(self, exposure: float | None) -> float:
        return self._readout_time

    async def prepare(self, trigger_info: TriggerInfo):
        if trigger_info.livetime is not None:
            await adcore.set_exposure_time_and_acquire_period_if_supplied(
                self, self._drv, trigger_info.livetime
            )
        await asyncio.gather(
            self._drv.trigger_mode.set(self._get_trigger_mode(trigger_info.trigger)),
            self._drv.num_images.set(
                999_999
                if trigger_info.total_number_of_triggers == 0
                else trigger_info.total_number_of_triggers
            ),
            self._drv.image_mode.set(adcore.ImageMode.multiple),
        )

    async def arm(self):
        # Standard arm the detector and wait for the acquire PV to be True
        self._arm_status = await adcore.start_acquiring_driver_and_ensure_status(
            self._drv
        )
        # The pilatus has an additional PV that goes True when the camserver
        # is actually ready. Should wait for that too or we risk dropping
        # a frame
        await wait_for_value(
            self._drv.armed,
            True,
            timeout=DEFAULT_TIMEOUT,
        )

    async def wait_for_idle(self):
        if self._arm_status:
            await self._arm_status

    @classmethod
    def _get_trigger_mode(cls, trigger: DetectorTrigger) -> PilatusTriggerMode:
        if trigger not in cls._supported_trigger_types.keys():
            raise ValueError(
                f"{cls.__name__} only supports the following trigger "
                f"types: {cls._supported_trigger_types.keys()} but was asked to "
                f"use {trigger}"
            )
        return cls._supported_trigger_types[trigger]

    async def disarm(self):
        await adcore.stop_busy_record(self._drv.acquire, False, timeout=1)



from ophyd_async.core import Device, StrictEnum
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw_rbv, epics_signal_w


class EigerTriggerMode(StrictEnum):
    internal = "ints"
    edge = "exts"
    gate = "exte"


class EigerDriverIO(Device):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.bit_depth = epics_signal_r(int, f"{prefix}BitDepthReadout")
        self.stale_parameters = epics_signal_r(bool, f"{prefix}StaleParameters")
        self.state = epics_signal_r(str, f"{prefix}DetectorState")
        self.roi_mode = epics_signal_rw_rbv(str, f"{prefix}RoiMode")

        self.acquire_time = epics_signal_rw_rbv(float, f"{prefix}CountTime")
        self.acquire_period = epics_signal_rw_rbv(float, f"{prefix}FrameTime")

        self.num_images = epics_signal_rw_rbv(int, f"{prefix}Nimages")
        self.num_triggers = epics_signal_rw_rbv(int, f"{prefix}Ntrigger")

        # TODO: Should be EigerTriggerMode enum, see https://github.com/DiamondLightSource/eiger-fastcs/issues/43
        self.trigger_mode = epics_signal_rw_rbv(str, f"{prefix}TriggerMode")

        self.arm = epics_signal_w(int, f"{prefix}Arm")
        self.disarm = epics_signal_w(int, f"{prefix}Disarm")
        self.abort = epics_signal_w(int, f"{prefix}Abort")

        self.beam_centre_x = epics_signal_rw_rbv(float, f"{prefix}BeamCenterX")
        self.beam_centre_y = epics_signal_rw_rbv(float, f"{prefix}BeamCenterY")

        self.det_distance = epics_signal_rw_rbv(float, f"{prefix}DetectorDistance")
        self.omega_start = epics_signal_rw_rbv(float, f"{prefix}OmegaStart")
        self.omega_increment = epics_signal_rw_rbv(float, f"{prefix}OmegaIncrement")

        self.photon_energy = epics_signal_rw_rbv(float, f"{prefix}PhotonEnergy")

        super().__init__(name)



from ophyd_async.core import Device, DeviceConnector
from ophyd_async.epics.core import PviDeviceConnector


def fastcs_connector(device: Device, uri: str) -> DeviceConnector:
    # TODO: add Tango support based on uri scheme
    connector = PviDeviceConnector(uri)
    connector.create_children_from_annotations(device)
    return connector



from collections.abc import Sequence

import numpy as np
import pytest
from pydantic import ValidationError

from ophyd_async.core import Array1D, Table


class MyTable(Table):
    bool: Array1D[np.bool_]
    uint: Array1D[np.uint32]
    str: Sequence[str]


@pytest.mark.parametrize(
    ["kwargs", "error_msg"],
    [
        (
            {"bool": [3, 4], "uint": [3, 4], "str": ["", ""]},
            "bool: Cannot cast [3, 4] to bool without losing precision",
        ),
        (
            {"bool": np.array([1], dtype=np.uint8), "uint": [-3], "str": [""]},
            "uint: Cannot cast [-3] to uint32 without losing precision",
        ),
        (
            {"bool": [0], "uint": np.array([1.8], dtype=np.float64), "str": [""]},
            "uint: Cannot cast [1.8] to uint32 without losing precision",
        ),
        (
            {"bool": [0, 1], "uint": [3, 4], "str": [44, ""]},
            "Input should be a valid string [type=string_type, input_value=44,",
        ),
    ],
)
def test_table_wrong_types(kwargs, error_msg):
    with pytest.raises(ValidationError) as cm:
        MyTable(**kwargs)
    assert error_msg in str(cm.value)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"bool": np.array([1], dtype=np.uint8), "uint": [3], "str": ["a"]},
        {"bool": [False], "uint": np.array([1], dtype=np.float64), "str": ["b"]},
    ],
)
def test_table_coerces(kwargs):
    t = MyTable(**kwargs)
    for k, v in t:
        assert v == pytest.approx(kwargs[k])



import asyncio
import traceback
from unittest.mock import Mock

import pytest

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    Device,
    DeviceCollector,
    DeviceVector,
    NotConnected,
    Reference,
    SignalRW,
    soft_signal_rw,
    wait_for_connection,
)
from ophyd_async.epics import motor
from ophyd_async.plan_stubs import ensure_connected


class DummyBaseDevice(Device):
    def __init__(self) -> None:
        self.connected = False
        super().__init__()

    async def connect(
        self, mock=False, timeout=DEFAULT_TIMEOUT, force_reconnect: bool = False
    ):
        self.connected = True


class DummyDeviceGroup(Device):
    def __init__(self, name: str) -> None:
        self.child1 = DummyBaseDevice()
        self._child2 = DummyBaseDevice()
        self.dict_with_children: DeviceVector[DummyBaseDevice] = DeviceVector(
            {123: DummyBaseDevice()}
        )
        super().__init__(name)


@pytest.fixture
def parent() -> DummyDeviceGroup:
    return DummyDeviceGroup("parent")


class DeviceWithNamedChild(Device):
    def __init__(self, name: str = "") -> None:
        super().__init__(name)
        self.child = soft_signal_rw(int, name="foo")


def test_device_signal_naming():
    device = DeviceWithNamedChild("bar")
    assert device.name == "bar"
    assert device.child.name == "foo"


class DeviceWithRefToSignal(Device):
    def __init__(self, signal: SignalRW[int]):
        self.signal_ref = Reference(signal)
        super().__init__(name="bat")

    def get_source(self) -> str:
        return self.signal_ref().source


def test_device_with_signal_ref_does_not_rename():
    device = DeviceWithNamedChild()
    device.set_name("bar")
    assert dict(device.children()) == {"child": device.child}
    private_device = DeviceWithRefToSignal(device.child)
    assert device.child.source == private_device.get_source()
    assert dict(private_device.children()) == {}
    assert device.name == "bar"
    assert device.child.name == "bar-child"
    assert private_device.name == "bat"


def test_device_children(parent: DummyDeviceGroup):
    names = ["child1", "_child2", "dict_with_children"]
    for idx, (name, child) in enumerate(parent.children()):
        assert name == names[idx]
        expected_type = (
            DeviceVector if name == "dict_with_children" else DummyBaseDevice
        )
        assert type(child) is expected_type
        assert child.parent == parent


def test_device_vector_children():
    parent = DummyDeviceGroup("root")

    device_vector_children = list(parent.dict_with_children.items())
    assert device_vector_children == [(123, parent.dict_with_children[123])]


async def test_children_of_device_have_set_names_and_get_connected(
    parent: DummyDeviceGroup,
):
    assert parent.name == "parent"
    assert parent.child1.name == "parent-child1"
    assert parent._child2.name == "parent-child2"
    assert parent.dict_with_children.name == "parent-dict_with_children"
    assert parent.dict_with_children[123].name == "parent-dict_with_children-123"

    await parent.connect()

    assert parent.child1.connected
    assert parent.dict_with_children[123].connected


async def test_device_with_device_collector():
    async with DeviceCollector(mock=True):
        parent = DummyDeviceGroup("parent")

    assert parent.name == "parent"
    assert parent.child1.name == "parent-child1"
    assert parent._child2.name == "parent-child2"
    assert parent.dict_with_children.name == "parent-dict_with_children"
    assert parent.dict_with_children[123].name == "parent-dict_with_children-123"
    assert parent.child1.connected
    assert parent.dict_with_children[123].connected


async def test_wait_for_connection():
    class DummyDeviceWithSleep(DummyBaseDevice):
        def __init__(self, name) -> None:
            self.set_name(name)

        async def connect(self, mock=False, timeout=DEFAULT_TIMEOUT):
            await asyncio.sleep(0.01)
            self.connected = True

    device1, device2 = DummyDeviceWithSleep("device1"), DummyDeviceWithSleep("device2")

    normal_coros = {"device1": device1.connect(), "device2": device2.connect()}

    await wait_for_connection(**normal_coros)

    assert device1.connected
    assert device2.connected


async def test_wait_for_connection_propagates_error(
    normal_coroutine, failing_coroutine
):
    failing_coros = {"test": normal_coroutine(), "failing": failing_coroutine()}

    with pytest.raises(NotConnected) as e:
        await wait_for_connection(**failing_coros)
        assert traceback.extract_tb(e.__traceback__)[-1].name == "failing_coroutine"


async def test_device_log_has_correct_name():
    device = DummyBaseDevice()
    assert device.log.extra["ophyd_async_device_name"] == ""
    device.set_name("device")
    assert device.log.extra["ophyd_async_device_name"] == "device"


class MotorBundle(Device):
    def __init__(self, name: str) -> None:
        self.X = motor.Motor("BLxxI-MO-TABLE-01:X")
        self.Y = motor.Motor("BLxxI-MO-TABLE-01:Y")
        self.V: DeviceVector[motor.Motor] = DeviceVector(
            {
                0: motor.Motor("BLxxI-MO-TABLE-21:X"),
                1: motor.Motor("BLxxI-MO-TABLE-21:Y"),
                2: motor.Motor("BLxxI-MO-TABLE-21:Z"),
            }
        )
        super().__init__(name)


async def test_device_with_children_lazily_connects(RE):
    parentMotor = MotorBundle("parentMotor")

    for device in [parentMotor, parentMotor.X, parentMotor.Y] + list(
        parentMotor.V.values()
    ):
        assert device._connect_task is None
    RE(ensure_connected(parentMotor, mock=True))

    for device in [parentMotor, parentMotor.X, parentMotor.Y] + list(
        parentMotor.V.values()
    ):
        assert (
            device._connect_task is not None
            and device._connect_task.done()
            and not device._connect_task.exception()
        )


@pytest.mark.parametrize("use_Mock", [False, True])
async def test_no_reconnect_signals_if_not_forced(use_Mock):
    parent = DummyDeviceGroup("parent")

    connect_mock_arg = Mock() if use_Mock else True

    async def inner_connect(mock, timeout, force_reconnect):
        parent.child1.connected = True

    parent.child1.connect = Mock(side_effect=inner_connect)
    await parent.connect(mock=connect_mock_arg, timeout=0.01)
    assert parent.child1.connected
    assert parent.child1.connect.call_count == 1
    await parent.connect(mock=connect_mock_arg, timeout=0.01)
    assert parent.child1.connected
    assert parent.child1.connect.call_count == 1

    for count in range(2, 10):
        await parent.connect(mock=connect_mock_arg, timeout=0.01, force_reconnect=True)
        assert parent.child1.connected
        assert parent.child1.connect.call_count == count



from __future__ import annotations

from ophyd_async.core import (
    StandardReadable,
)
from ophyd_async.tango.base_devices._base_device import TangoDevice
from tango import DeviceProxy


class TangoReadable(TangoDevice, StandardReadable):
    """
    General class for readable TangoDevices. Extends StandardReadable to provide
    attributes for Tango devices.

    Usage: to proper signals mount should be awaited:
    new_device = await TangoDevice(<tango_device>)

    Attributes
    ----------
    trl : str
        Tango resource locator, typically of the device server.
    proxy : AsyncDeviceProxy
        AsyncDeviceProxy object for the device. This is created when the
        device is connected.
    """

    def __init__(
        self,
        trl: str | None = None,
        device_proxy: DeviceProxy | None = None,
        name: str = "",
    ) -> None:
        TangoDevice.__init__(self, trl, device_proxy=device_proxy, name=name)



import asyncio

from pydantic import BaseModel, Field

from ophyd_async.core import FlyerController, wait_for_value

from ._block import BitMux, PcompBlock, PcompDirection, SeqBlock, TimeUnits
from ._table import SeqTable


class SeqTableInfo(BaseModel):
    sequence_table: SeqTable = Field(strict=True)
    repeats: int = Field(ge=0)
    prescale_as_us: float = Field(default=1, ge=0)  # microseconds


class StaticSeqTableTriggerLogic(FlyerController[SeqTableInfo]):
    def __init__(self, seq: SeqBlock) -> None:
        self.seq = seq

    async def prepare(self, value: SeqTableInfo):
        await asyncio.gather(
            self.seq.prescale_units.set(TimeUnits.us),
            self.seq.enable.set(BitMux.zero),
        )
        await asyncio.gather(
            self.seq.prescale.set(value.prescale_as_us),
            self.seq.repeats.set(value.repeats),
            self.seq.table.set(value.sequence_table),
        )

    async def kickoff(self) -> None:
        await self.seq.enable.set(BitMux.one)
        await wait_for_value(self.seq.active, True, timeout=1)

    async def complete(self) -> None:
        await wait_for_value(self.seq.active, False, timeout=None)

    async def stop(self):
        await self.seq.enable.set(BitMux.zero)
        await wait_for_value(self.seq.active, False, timeout=1)


class PcompInfo(BaseModel):
    start_postion: int = Field(description="start position in counts")
    pulse_width: int = Field(description="width of a single pulse in counts", gt=0)
    rising_edge_step: int = Field(
        description="step between rising edges of pulses in counts", gt=0
    )  #
    number_of_pulses: int = Field(
        description=(
            "Number of pulses to send before the PCOMP block is disarmed. "
            "0 means infinite."
        ),
        ge=0,
    )
    direction: PcompDirection = Field(
        description=(
            "Specifies which direction the motor counts should be "
            "moving. Pulses won't be sent unless the values are moving in "
            "this direction."
        )
    )


class StaticPcompTriggerLogic(FlyerController[PcompInfo]):
    def __init__(self, pcomp: PcompBlock) -> None:
        self.pcomp = pcomp

    async def prepare(self, value: PcompInfo):
        await self.pcomp.enable.set(BitMux.zero)
        await asyncio.gather(
            self.pcomp.start.set(value.start_postion),
            self.pcomp.width.set(value.pulse_width),
            self.pcomp.step.set(value.rising_edge_step),
            self.pcomp.pulses.set(value.number_of_pulses),
            self.pcomp.dir.set(value.direction),
        )

    async def kickoff(self) -> None:
        await self.pcomp.enable.set(BitMux.one)
        await wait_for_value(self.pcomp.active, True, timeout=1)

    async def complete(self, timeout: float | None = None) -> None:
        await wait_for_value(self.pcomp.active, False, timeout=timeout)

    async def stop(self):
        await self.pcomp.enable.set(BitMux.zero)
        await wait_for_value(self.pcomp.active, False, timeout=1)



from collections.abc import AsyncGenerator, AsyncIterator
from pathlib import Path

import h5py
import numpy as np
from bluesky.protocols import StreamAsset
from event_model import DataKey

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    HDFDataset,
    HDFFile,
    PathProvider,
    observe_value,
    soft_signal_r_and_setter,
)

# raw data path
DATA_PATH = "/entry/data/data"

# pixel sum path
SUM_PATH = "/entry/sum"

MAX_UINT8_VALUE = np.iinfo(np.uint8).max


def generate_gaussian_blob(height: int, width: int) -> np.ndarray:
    """Make a Gaussian Blob with float values in range 0..1"""
    x, y = np.meshgrid(np.linspace(-1, 1, width), np.linspace(-1, 1, height))
    d = np.sqrt(x * x + y * y)
    blob = np.exp(-(d**2))
    return blob


def generate_interesting_pattern(x: float, y: float) -> float:
    """This function is interesting in x and y in range -10..10, returning
    a float value in range 0..1
    """
    z = 0.5 + (np.sin(x) ** 10 + np.cos(10 + y * x) * np.cos(x)) / 2
    return z


class PatternGenerator:
    def __init__(
        self,
        saturation_exposure_time: float = 0.1,
        detector_width: int = 320,
        detector_height: int = 240,
    ) -> None:
        self.saturation_exposure_time = saturation_exposure_time
        self.exposure = saturation_exposure_time
        self.x = 0.0
        self.y = 0.0
        self.height = detector_height
        self.width = detector_width
        self.image_counter: int = 0

        # it automatically initializes to 0
        self.counter_signal, self._set_counter_signal = soft_signal_r_and_setter(int)
        self._full_intensity_blob = (
            generate_gaussian_blob(width=detector_width, height=detector_height)
            * MAX_UINT8_VALUE
        )
        self._hdf_stream_provider: HDFFile | None = None
        self._handle_for_h5_file: h5py.File | None = None
        self.target_path: Path | None = None

    def write_data_to_dataset(self, path: str, data_shape: tuple[int, ...], data):
        """Write data to named dataset, resizing to fit and flushing after."""
        assert self._handle_for_h5_file, "no file has been opened!"
        dset = self._handle_for_h5_file[path]
        assert isinstance(
            dset, h5py.Dataset
        ), f"Expected {path} to be dataset, got {dset}"
        dset.resize((self.image_counter + 1,) + data_shape)
        dset[self.image_counter] = data
        dset.flush()

    async def write_image_to_file(self) -> None:
        # generate the simulated data
        intensity: float = generate_interesting_pattern(self.x, self.y)
        detector_data = (
            self._full_intensity_blob
            * intensity
            * self.exposure
            / self.saturation_exposure_time
        ).astype(np.uint8)

        # Write the data and sum
        self.write_data_to_dataset(DATA_PATH, (self.height, self.width), detector_data)
        self.write_data_to_dataset(SUM_PATH, (), np.sum(detector_data))

        # counter increment is last
        # as only at this point the new data is visible from the outside
        self.image_counter += 1
        self._set_counter_signal(self.image_counter)

    def set_exposure(self, value: float) -> None:
        self.exposure = value

    def set_x(self, value: float) -> None:
        self.x = value

    def set_y(self, value: float) -> None:
        self.y = value

    async def open_file(
        self, path_provider: PathProvider, name: str, multiplier: int = 1
    ) -> dict[str, DataKey]:
        await self.counter_signal.connect()

        self.target_path = self._get_new_path(path_provider)
        self._path_provider = path_provider

        self._handle_for_h5_file = h5py.File(self.target_path, "w", libver="latest")

        assert self._handle_for_h5_file, "not loaded the file right"

        self._handle_for_h5_file.create_dataset(
            name=DATA_PATH,
            shape=(0, self.height, self.width),
            dtype=np.uint8,
            maxshape=(None, self.height, self.width),
        )
        self._handle_for_h5_file.create_dataset(
            name=SUM_PATH,
            shape=(0,),
            dtype=np.float64,
            maxshape=(None,),
        )

        # once datasets written, can switch the model to single writer multiple reader
        self._handle_for_h5_file.swmr_mode = True
        self.multiplier = multiplier

        outer_shape = (multiplier,) if multiplier > 1 else ()

        # cache state to self
        # Add the main data
        self._datasets = [
            HDFDataset(
                data_key=name,
                dataset=DATA_PATH,
                shape=(self.height, self.width),
                multiplier=multiplier,
            ),
            HDFDataset(
                f"{name}-sum",
                dataset=SUM_PATH,
                shape=(),
                multiplier=multiplier,
            ),
        ]

        describe = {
            ds.data_key: DataKey(
                source="sim://pattern-generator-hdf-file",
                shape=list(outer_shape) + list(ds.shape),
                dtype="array" if ds.shape else "number",
                external="STREAM:",
            )
            for ds in self._datasets
        }
        return describe

    def _get_new_path(self, path_provider: PathProvider) -> Path:
        info = path_provider(device_name="pattern")
        new_path: Path = info.directory_path / info.filename
        return new_path

    async def collect_stream_docs(
        self, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        """
        stream resource says "here is a dataset",
        stream datum says "here are N frames in that stream resource",
        you get one stream resource and many stream datums per scan
        """
        if self._handle_for_h5_file:
            self._handle_for_h5_file.flush()
        # when already something was written to the file
        if indices_written:
            # if no frames arrived yet, there's no file to speak of
            # cannot get the full filename the HDF writer will write
            # until the first frame comes in
            if not self._hdf_stream_provider:
                assert self.target_path, "open file has not been called"
                self._hdf_stream_provider = HDFFile(
                    self.target_path,
                    self._datasets,
                )
                for doc in self._hdf_stream_provider.stream_resources():
                    yield "stream_resource", doc
            if self._hdf_stream_provider:
                for doc in self._hdf_stream_provider.stream_data(indices_written):
                    yield "stream_datum", doc

    def close(self) -> None:
        if self._handle_for_h5_file:
            self._handle_for_h5_file.close()
            self._handle_for_h5_file = None

    async def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        async for num_captured in observe_value(self.counter_signal, timeout=timeout):
            yield num_captured // self.multiplier



import asyncio
import functools
import time
from abc import abstractmethod
from collections.abc import Callable, Coroutine
from enum import Enum
from typing import Any, TypeVar, cast

import numpy as np
from bluesky.protocols import Descriptor, Reading

from ophyd_async.core import (
    AsyncStatus,
    Callback,
    NotConnected,
    SignalBackend,
    SignalDatatypeT,
    get_dtype,
    get_unique,
    wait_for_connection,
)
from tango import (
    AttrDataFormat,
    AttributeInfoEx,
    CmdArgType,
    CommandInfo,
    DevFailed,  # type: ignore
    DeviceProxy,
    DevState,
    EventType,
)
from tango.asyncio import DeviceProxy as AsyncDeviceProxy
from tango.asyncio_executor import (
    AsyncioExecutor,
    get_global_executor,
    set_global_executor,
)
from tango.utils import is_array, is_binary, is_bool, is_float, is_int, is_str

# time constant to wait for timeout
A_BIT = 1e-5

R = TypeVar("R")


def ensure_proper_executor(
    func: Callable[..., Coroutine[Any, Any, R]],
) -> Callable[..., Coroutine[Any, Any, R]]:
    @functools.wraps(func)
    async def wrapper(self: Any, *args: Any, **kwargs: Any) -> R:
        current_executor: AsyncioExecutor = get_global_executor()  # type: ignore
        if not current_executor.in_executor_context():  # type: ignore
            set_global_executor(AsyncioExecutor())
        return await func(self, *args, **kwargs)

    return cast(Callable[..., Coroutine[Any, Any, R]], wrapper)


def get_python_type(tango_type: CmdArgType) -> tuple[bool, object, str]:
    array = is_array(tango_type)
    if is_int(tango_type, True):
        return array, int, "integer"
    if is_float(tango_type, True):
        return array, float, "number"
    if is_bool(tango_type, True):
        return array, bool, "integer"
    if is_str(tango_type, True):
        return array, str, "string"
    if is_binary(tango_type, True):
        return array, list[str], "string"
    if tango_type == CmdArgType.DevEnum:
        return array, Enum, "string"
    if tango_type == CmdArgType.DevState:
        return array, CmdArgType.DevState, "string"
    if tango_type == CmdArgType.DevUChar:
        return array, int, "integer"
    if tango_type == CmdArgType.DevVoid:
        return array, None, "string"
    raise TypeError("Unknown TangoType")


class TangoProxy:
    support_events: bool = True
    _proxy: DeviceProxy
    _name: str

    def __init__(self, device_proxy: DeviceProxy, name: str):
        self._proxy = device_proxy
        self._name = name

    async def connect(self) -> None:
        """perform actions after proxy is connected, e.g. checks if signal
        can be subscribed"""

    @abstractmethod
    async def get(self) -> object:
        """Get value from TRL"""

    @abstractmethod
    async def get_w_value(self) -> object:
        """Get last written value from TRL"""

    @abstractmethod
    async def put(
        self, value: object | None, wait: bool = True, timeout: float | None = None
    ) -> AsyncStatus | None:
        """Put value to TRL"""

    @abstractmethod
    async def get_config(self) -> AttributeInfoEx | CommandInfo:
        """Get TRL config async"""

    @abstractmethod
    async def get_reading(self) -> Reading:
        """Get reading from TRL"""

    @abstractmethod
    def has_subscription(self) -> bool:
        """indicates, that this trl already subscribed"""

    @abstractmethod
    def subscribe_callback(self, callback: Callback | None):
        """subscribe tango CHANGE event to callback"""

    @abstractmethod
    def unsubscribe_callback(self):
        """delete CHANGE event subscription"""

    @abstractmethod
    def set_polling(
        self,
        allow_polling: bool = True,
        polling_period: float = 0.1,
        abs_change=None,
        rel_change=None,
    ):
        """Set polling parameters"""


class AttributeProxy(TangoProxy):
    _callback: Callback | None = None
    _eid: int | None = None
    _poll_task: asyncio.Task | None = None
    _abs_change: float | None = None
    _rel_change: float | None = 0.1
    _polling_period: float = 0.1
    _allow_polling: bool = False
    exception: BaseException | None = None
    _last_reading: Reading = Reading(value=None, timestamp=0, alarm_severity=0)

    async def connect(self) -> None:
        try:
            # I have to typehint proxy as tango.DeviceProxy because
            # tango.asyncio.DeviceProxy cannot be used as a typehint.
            # This means pyright will not be able to see that
            # subscribe_event is awaitable.
            eid = await self._proxy.subscribe_event(  # type: ignore
                self._name, EventType.CHANGE_EVENT, self._event_processor
            )
            await self._proxy.unsubscribe_event(eid)
            self.support_events = True
        except Exception:
            pass

    @ensure_proper_executor
    async def get(self) -> Coroutine[Any, Any, object]:
        attr = await self._proxy.read_attribute(self._name)
        return attr.value

    @ensure_proper_executor
    async def get_w_value(self) -> object:
        attr = await self._proxy.read_attribute(self._name)
        return attr.w_value

    @ensure_proper_executor
    async def put(
        self, value: object | None, wait: bool = True, timeout: float | None = None
    ) -> AsyncStatus | None:
        # TODO: remove the timeout from this as it is handled at the signal level
        if wait:
            try:

                async def _write():
                    return await self._proxy.write_attribute(self._name, value)

                task = asyncio.create_task(_write())
                await asyncio.wait_for(task, timeout)
            except asyncio.TimeoutError as te:
                raise TimeoutError(f"{self._name} attr put failed: Timeout") from te
            except DevFailed as de:
                raise RuntimeError(
                    f"{self._name} device" f" failure: {de.args[0].desc}"
                ) from de

        else:
            rid = await self._proxy.write_attribute_asynch(self._name, value)

            async def wait_for_reply(rd: int, to: float | None):
                start_time = time.time()
                while True:
                    try:
                        # I have to typehint proxy as tango.DeviceProxy because
                        # tango.asyncio.DeviceProxy cannot be used as a typehint.
                        # This means pyright will not be able to see that
                        # write_attribute_reply is awaitable.
                        await self._proxy.write_attribute_reply(rd)  # type: ignore
                        break
                    except DevFailed as exc:
                        if exc.args[0].reason == "API_AsynReplyNotArrived":
                            await asyncio.sleep(A_BIT)
                            if to and (time.time() - start_time > to):
                                raise TimeoutError(
                                    f"{self._name} attr put failed:" f" Timeout"
                                ) from exc
                        else:
                            raise RuntimeError(
                                f"{self._name} device failure:" f" {exc.args[0].desc}"
                            ) from exc

            return AsyncStatus(wait_for_reply(rid, timeout))

    @ensure_proper_executor
    async def get_config(self) -> AttributeInfoEx:
        return await self._proxy.get_attribute_config(self._name)

    @ensure_proper_executor
    async def get_reading(self) -> Reading:
        attr = await self._proxy.read_attribute(self._name)
        reading = Reading(
            value=attr.value, timestamp=attr.time.totime(), alarm_severity=attr.quality
        )
        self._last_reading = reading
        return reading

    def has_subscription(self) -> bool:
        return bool(self._callback)

    def subscribe_callback(self, callback: Callback | None):
        # If the attribute supports events, then we can subscribe to them
        # If the callback is not a callable, then we raise an error
        if callback is not None and not callable(callback):
            raise RuntimeError("Callback must be a callable")

        self._callback = callback
        if self.support_events:
            """add user callback to CHANGE event subscription"""
            if not self._eid:
                self._eid = self._proxy.subscribe_event(
                    self._name,
                    EventType.CHANGE_EVENT,
                    self._event_processor,
                    green_mode=False,
                )
        elif self._allow_polling:
            """start polling if no events supported"""
            if self._callback is not None:

                async def _poll():
                    while True:
                        try:
                            await self.poll()
                        except RuntimeError as exc:
                            self.exception = exc
                            await asyncio.sleep(1)

                self._poll_task = asyncio.create_task(_poll())
        else:
            self.unsubscribe_callback()
            raise RuntimeError(
                f"Cannot set event for {self._name}. "
                "Cannot set a callback on an attribute that does not support events and"
                " for which polling is disabled."
            )

    def unsubscribe_callback(self):
        if self._eid:
            self._proxy.unsubscribe_event(self._eid, green_mode=False)
            self._eid = None
        if self._poll_task:
            self._poll_task.cancel()
            self._poll_task = None
            if self._callback is not None:
                # Call the callback with the last reading
                try:
                    self._callback(self._last_reading)
                except TypeError:
                    pass
        self._callback = None

    def _event_processor(self, event):
        if not event.err:
            reading = Reading(
                value=event.attr_value.value,
                timestamp=event.get_date().totime(),
                alarm_severity=event.attr_value.quality,
            )
            if self._callback is not None:
                self._callback(reading)

    async def poll(self):
        """
        Poll the attribute and call the callback if the value has changed by more
        than the absolute or relative change. This function is used when an attribute
        that does not support events is cached or a callback is passed to it.
        """
        try:
            last_reading = await self.get_reading()
            flag = 0
            # Initial reading
            if self._callback is not None:
                self._callback(last_reading)
        except Exception as e:
            raise RuntimeError(f"Could not poll the attribute: {e}") from e

        try:
            # If the value is a number, we can check for changes
            if isinstance(last_reading["value"], int | float):
                while True:
                    await asyncio.sleep(self._polling_period)
                    reading = await self.get_reading()
                    if reading is None or reading["value"] is None:
                        continue
                    diff = abs(reading["value"] - last_reading["value"])
                    if self._abs_change is not None and diff >= abs(self._abs_change):
                        if self._callback is not None:
                            self._callback(reading)
                            flag = 0

                    elif (
                        self._rel_change is not None
                        and diff >= self._rel_change * abs(last_reading["value"])
                    ):
                        if self._callback is not None:
                            self._callback(reading)
                            flag = 0

                    else:
                        flag = (flag + 1) % 4
                        if flag == 0 and self._callback is not None:
                            self._callback(reading)

                    last_reading = reading.copy()
                    if self._callback is None:
                        break
            # If the value is not a number, we can only poll
            else:
                while True:
                    await asyncio.sleep(self._polling_period)
                    flag = (flag + 1) % 4
                    if flag == 0:
                        reading = await self.get_reading()
                        if reading is None or reading["value"] is None:
                            continue
                        if isinstance(reading["value"], np.ndarray):
                            if not np.array_equal(
                                reading["value"], last_reading["value"]
                            ):
                                if self._callback is not None:
                                    self._callback(reading)
                                else:
                                    break
                        else:
                            if reading["value"] != last_reading["value"]:
                                if self._callback is not None:
                                    self._callback(reading)
                                else:
                                    break
                        last_reading = reading.copy()
        except Exception as e:
            raise RuntimeError(f"Could not poll the attribute: {e}") from e

    def set_polling(
        self,
        allow_polling: bool = False,
        polling_period: float = 0.5,
        abs_change: float | None = None,
        rel_change: float | None = 0.1,
    ):
        """
        Set the polling parameters.
        """
        self._allow_polling = allow_polling
        self._polling_period = polling_period
        self._abs_change = abs_change
        self._rel_change = rel_change


class CommandProxy(TangoProxy):
    _last_reading: Reading = Reading(value=None, timestamp=0, alarm_severity=0)

    def subscribe_callback(self, callback: Callback | None) -> None:
        raise NotImplementedError("Cannot subscribe to commands")

    def unsubscribe_callback(self) -> None:
        raise NotImplementedError("Cannot unsubscribe from commands")

    async def get(self) -> object:
        return self._last_reading["value"]

    async def get_w_value(self) -> object:
        return self._last_reading["value"]

    async def connect(self) -> None:
        pass

    @ensure_proper_executor
    async def put(
        self, value: object | None, wait: bool = True, timeout: float | None = None
    ) -> AsyncStatus | None:
        if wait:
            try:

                async def _put():
                    return await self._proxy.command_inout(self._name, value)

                task = asyncio.create_task(_put())
                val = await asyncio.wait_for(task, timeout)
                self._last_reading = Reading(
                    value=val, timestamp=time.time(), alarm_severity=0
                )
            except asyncio.TimeoutError as te:
                raise TimeoutError(f"{self._name} command failed: Timeout") from te
            except DevFailed as de:
                raise RuntimeError(
                    f"{self._name} device" f" failure: {de.args[0].desc}"
                ) from de

        else:
            rid = self._proxy.command_inout_asynch(self._name, value)

            async def wait_for_reply(rd: int, to: float | None):
                start_time = time.time()
                while True:
                    try:
                        reply_value = self._proxy.command_inout_reply(rd)
                        self._last_reading = Reading(
                            value=reply_value, timestamp=time.time(), alarm_severity=0
                        )
                        break
                    except DevFailed as de_exc:
                        if de_exc.args[0].reason == "API_AsynReplyNotArrived":
                            await asyncio.sleep(A_BIT)
                            if to and time.time() - start_time > to:
                                raise TimeoutError(
                                    "Timeout while waiting for command reply"
                                ) from de_exc
                        else:
                            raise RuntimeError(
                                f"{self._name} device failure:"
                                f" {de_exc.args[0].desc}"
                            ) from de_exc

            return AsyncStatus(wait_for_reply(rid, timeout))

    @ensure_proper_executor
    async def get_config(self) -> CommandInfo:
        return await self._proxy.get_command_config(self._name)

    async def get_reading(self) -> Reading:
        reading = Reading(
            value=self._last_reading["value"],
            timestamp=self._last_reading["timestamp"],
            alarm_severity=self._last_reading.get("alarm_severity", 0),
        )
        return reading

    def set_polling(
        self,
        allow_polling: bool = False,
        polling_period: float = 0.5,
        abs_change: float | None = None,
        rel_change: float | None = 0.1,
    ):
        pass


def get_dtype_extended(datatype) -> object | None:
    # DevState tango type does not have numpy equivalents
    dtype = get_dtype(datatype)
    if dtype == np.object_:
        if datatype.__args__[1].__args__[0] == DevState:
            dtype = CmdArgType.DevState
    return dtype


def get_trl_descriptor(
    datatype: type | None,
    tango_resource: str,
    tr_configs: dict[str, AttributeInfoEx | CommandInfo],
) -> Descriptor:
    tr_dtype = {}
    for tr_name, config in tr_configs.items():
        if isinstance(config, AttributeInfoEx):
            _, dtype, descr = get_python_type(config.data_type)
            tr_dtype[tr_name] = config.data_format, dtype, descr
        elif isinstance(config, CommandInfo):
            if (
                config.in_type != CmdArgType.DevVoid
                and config.out_type != CmdArgType.DevVoid
                and config.in_type != config.out_type
            ):
                raise RuntimeError(
                    "Commands with different in and out dtypes are not supported"
                )
            array, dtype, descr = get_python_type(
                config.in_type
                if config.in_type != CmdArgType.DevVoid
                else config.out_type
            )
            tr_dtype[tr_name] = (
                AttrDataFormat.SPECTRUM if array else AttrDataFormat.SCALAR,
                dtype,
                descr,
            )
        else:
            raise RuntimeError(f"Unknown config type: {type(config)}")
    tr_format, tr_dtype, tr_dtype_desc = get_unique(tr_dtype, "typeids")

    # tango commands are limited in functionality:
    # they do not have info about shape and Enum labels
    trl_config = list(tr_configs.values())[0]
    max_x: int = (
        trl_config.max_dim_x
        if hasattr(trl_config, "max_dim_x")
        else np.iinfo(np.int32).max
    )
    max_y: int = (
        trl_config.max_dim_y
        if hasattr(trl_config, "max_dim_y")
        else np.iinfo(np.int32).max
    )
    # is_attr = hasattr(trl_config, "enum_labels")
    # trl_choices = list(trl_config.enum_labels) if is_attr else []

    if tr_format in [AttrDataFormat.SPECTRUM, AttrDataFormat.IMAGE]:
        # This is an array
        if datatype:
            # Check we wanted an array of this type
            dtype = get_dtype_extended(datatype)
            if not dtype:
                raise TypeError(
                    f"{tango_resource} has type [{tr_dtype}] not {datatype.__name__}"
                )
            if dtype != tr_dtype:
                raise TypeError(f"{tango_resource} has type [{tr_dtype}] not [{dtype}]")

        if tr_format == AttrDataFormat.SPECTRUM:
            return Descriptor(source=tango_resource, dtype="array", shape=[max_x])
        elif tr_format == AttrDataFormat.IMAGE:
            return Descriptor(
                source=tango_resource, dtype="array", shape=[max_y, max_x]
            )

    else:
        if tr_dtype in (Enum, CmdArgType.DevState):
            # if tr_dtype == CmdArgType.DevState:
            #     trl_choices = list(DevState.names.keys())

            if datatype:
                if not issubclass(datatype, Enum | DevState):
                    raise TypeError(
                        f"{tango_resource} has type Enum not {datatype.__name__}"
                    )
                # if tr_dtype == Enum and is_attr:
                #     if isinstance(datatype, DevState):
                #         choices = tuple(v.name for v in datatype)
                #         if set(choices) != set(trl_choices):
                #             raise TypeError(
                #                 f"{tango_resource} has choices {trl_choices} "
                #                 f"not {choices}"
                #             )
            return Descriptor(source=tango_resource, dtype="string", shape=[])
        else:
            if datatype and not issubclass(tr_dtype, datatype):
                raise TypeError(
                    f"{tango_resource} has type {tr_dtype.__name__} "
                    f"not {datatype.__name__}"
                )
            return Descriptor(source=tango_resource, dtype=tr_dtype_desc, shape=[])

    raise RuntimeError(f"Error getting descriptor for {tango_resource}")


async def get_tango_trl(
    full_trl: str, device_proxy: DeviceProxy | TangoProxy | None, timeout: float
) -> TangoProxy:
    if isinstance(device_proxy, TangoProxy):
        return device_proxy
    device_trl, trl_name = full_trl.rsplit("/", 1)
    trl_name = trl_name.lower()
    if device_proxy is None:
        device_proxy = await AsyncDeviceProxy(device_trl, timeout=timeout)

    # all attributes can be always accessible with low register
    if isinstance(device_proxy, DeviceProxy):
        all_attrs = [
            attr_name.lower() for attr_name in device_proxy.get_attribute_list()
        ]
    else:
        raise TypeError(
            f"device_proxy must be an instance of DeviceProxy for {full_trl}"
        )
    if trl_name in all_attrs:
        return AttributeProxy(device_proxy, trl_name)

    # all commands can be always accessible with low register
    all_cmds = [cmd_name.lower() for cmd_name in device_proxy.get_command_list()]
    if trl_name in all_cmds:
        return CommandProxy(device_proxy, trl_name)

    # If version is below tango 9, then pipes are not supported
    if device_proxy.info().server_version >= 9:
        # all pipes can be always accessible with low register
        all_pipes = [pipe_name.lower() for pipe_name in device_proxy.get_pipe_list()]
        if trl_name in all_pipes:
            raise NotImplementedError("Pipes are not supported")

    raise RuntimeError(f"{trl_name} cannot be found in {device_proxy.name()}")


class TangoSignalBackend(SignalBackend[SignalDatatypeT]):
    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        read_trl: str = "",
        write_trl: str = "",
        device_proxy: DeviceProxy | None = None,
    ):
        self.device_proxy = device_proxy
        self.read_trl = read_trl
        self.write_trl = write_trl
        self.proxies: dict[str, TangoProxy | DeviceProxy | None] = {
            read_trl: self.device_proxy,
            write_trl: self.device_proxy,
        }
        self.trl_configs: dict[str, AttributeInfoEx] = {}
        self.descriptor: Descriptor = {}  # type: ignore
        self._polling: tuple[bool, float, float | None, float | None] = (
            False,
            0.1,
            None,
            0.1,
        )
        self.support_events: bool = True
        self.status: AsyncStatus | None = None
        super().__init__(datatype)

    @classmethod
    def datatype_allowed(cls, dtype: Any) -> bool:
        return dtype in (int, float, str, bool, np.ndarray, Enum, DevState)

    def set_trl(self, read_trl: str = "", write_trl: str = ""):
        self.read_trl = read_trl
        self.write_trl = write_trl if write_trl else read_trl
        self.proxies = {
            read_trl: self.device_proxy,
            write_trl: self.device_proxy,
        }

    def source(self, name: str, read: bool) -> str:
        return self.read_trl if read else self.write_trl

    async def _connect_and_store_config(self, trl: str, timeout: float) -> None:
        if not trl:
            raise RuntimeError(f"trl not set for {self}")
        try:
            self.proxies[trl] = await get_tango_trl(trl, self.proxies[trl], timeout)
            if self.proxies[trl] is None:
                raise NotConnected(f"Not connected to {trl}")
            # Pyright does not believe that self.proxies[trl] is not None despite
            # the check above
            await self.proxies[trl].connect()  # type: ignore
            self.trl_configs[trl] = await self.proxies[trl].get_config()  # type: ignore
            self.proxies[trl].support_events = self.support_events  # type: ignore
        except TimeoutError as ce:
            raise NotConnected(f"tango://{trl}") from ce

    async def connect(self, timeout: float) -> None:
        if not self.read_trl:
            raise RuntimeError(f"trl not set for {self}")
        if self.read_trl != self.write_trl:
            # Different, need to connect both
            await wait_for_connection(
                read_trl=self._connect_and_store_config(self.read_trl, timeout),
                write_trl=self._connect_and_store_config(self.write_trl, timeout),
            )
        else:
            # The same, so only need to connect one
            await self._connect_and_store_config(self.read_trl, timeout)
        self.proxies[self.read_trl].set_polling(*self._polling)  # type: ignore
        self.descriptor = get_trl_descriptor(
            self.datatype, self.read_trl, self.trl_configs
        )

    async def put(self, value: SignalDatatypeT | None, wait=True, timeout=None) -> None:
        if self.proxies[self.write_trl] is None:
            raise NotConnected(f"Not connected to {self.write_trl}")
        self.status = None
        put_status = await self.proxies[self.write_trl].put(value, wait, timeout)  # type: ignore
        self.status = put_status

    async def get_datakey(self, source: str) -> Descriptor:
        return self.descriptor

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        if self.proxies[self.read_trl] is None:
            raise NotConnected(f"Not connected to {self.read_trl}")
        return await self.proxies[self.read_trl].get_reading()  # type: ignore

    async def get_value(self) -> SignalDatatypeT:
        if self.proxies[self.read_trl] is None:
            raise NotConnected(f"Not connected to {self.read_trl}")
        proxy = self.proxies[self.read_trl]
        if proxy is None:
            raise NotConnected(f"Not connected to {self.read_trl}")
        return cast(SignalDatatypeT, await proxy.get())

    async def get_setpoint(self) -> SignalDatatypeT:
        if self.proxies[self.write_trl] is None:
            raise NotConnected(f"Not connected to {self.write_trl}")
        proxy = self.proxies[self.write_trl]
        if proxy is None:
            raise NotConnected(f"Not connected to {self.write_trl}")
        return cast(SignalDatatypeT, await proxy.get_w_value())

    def set_callback(self, callback: Callback | None) -> None:
        if self.proxies[self.read_trl] is None:
            raise NotConnected(f"Not connected to {self.read_trl}")
        if self.support_events is False and self._polling[0] is False:
            raise RuntimeError(
                f"Cannot set event for {self.read_trl}. "
                "Cannot set a callback on an attribute that does not support events and"
                " for which polling is disabled."
            )

        if callback:
            try:
                assert not self.proxies[self.read_trl].has_subscription()  # type: ignore
                self.proxies[self.read_trl].subscribe_callback(callback)  # type: ignore
            except AssertionError as ae:
                raise RuntimeError(
                    "Cannot set a callback when one" " is already set"
                ) from ae
            except RuntimeError as exc:
                raise RuntimeError(
                    f"Cannot set callback" f" for {self.read_trl}. {exc}"
                ) from exc

        else:
            self.proxies[self.read_trl].unsubscribe_callback()  # type: ignore

    def set_polling(
        self,
        allow_polling: bool = True,
        polling_period: float = 0.1,
        abs_change: float | None = None,
        rel_change: float | None = 0.1,
    ):
        self._polling = (allow_polling, polling_period, abs_change, rel_change)

    def allow_events(self, allow: bool = True):
        self.support_events = allow



import asyncio
import logging
import re
import time
from asyncio import Event
from unittest.mock import ANY, Mock

import pytest
from bluesky.protocols import Reading

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DeviceCollector,
    MockSignalBackend,
    NotConnected,
    Signal,
    SignalR,
    SignalRW,
    SoftSignalBackend,
    StandardReadable,
    assert_configuration,
    assert_reading,
    assert_value,
    callback_on_mock_put,
    set_and_wait_for_other_value,
    set_and_wait_for_value,
    set_mock_put_proceeds,
    set_mock_value,
    soft_signal_r_and_setter,
    soft_signal_rw,
    wait_for_value,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw
from ophyd_async.plan_stubs import ensure_connected


def num_occurrences(substring: str, string: str) -> int:
    return len(list(re.finditer(re.escape(substring), string)))


async def test_signal_connects_to_previous_backend(caplog):
    caplog.set_level(logging.DEBUG)
    int_mock_backend = MockSignalBackend(SoftSignalBackend(int), Mock())
    original_connect = int_mock_backend.connect
    times_backend_connect_called = 0

    async def new_connect(timeout=1):
        nonlocal times_backend_connect_called
        times_backend_connect_called += 1
        await asyncio.sleep(0.1)
        await original_connect(timeout=timeout)

    int_mock_backend.connect = new_connect
    signal = Signal(int_mock_backend)
    await asyncio.gather(signal.connect(), signal.connect())
    assert num_occurrences(f"Connecting to {signal.source}", caplog.text) == 1
    assert times_backend_connect_called == 1


async def test_signal_connects_with_force_reconnect(caplog):
    caplog.set_level(logging.DEBUG)
    signal = Signal(MockSignalBackend(SoftSignalBackend(int), Mock()))
    await signal.connect()
    assert num_occurrences(f"Connecting to {signal.source}", caplog.text) == 1
    await signal.connect(force_reconnect=True)
    assert num_occurrences(f"Connecting to {signal.source}", caplog.text) == 2


async def test_signal_lazily_connects(RE):
    class MockSignalBackendFailingFirst(MockSignalBackend):
        succeed_on_connect = False

        async def connect(self, timeout=DEFAULT_TIMEOUT):
            if self.succeed_on_connect:
                self.succeed_on_connect = False
                await super().connect(timeout=timeout)
            else:
                self.succeed_on_connect = True
                raise RuntimeError("connect fail")

    signal = SignalRW(MockSignalBackendFailingFirst(SoftSignalBackend(int), Mock()))

    with pytest.raises(RuntimeError, match="connect fail"):
        await signal.connect(mock=False)

    assert (
        signal._connect_task
        and signal._connect_task.done()
        and signal._connect_task.exception()
    )

    RE(ensure_connected(signal, mock=False))
    assert (
        signal._connect_task
        and signal._connect_task.done()
        and not signal._connect_task.exception()
    )

    with pytest.raises(NotConnected, match="RuntimeError: connect fail"):
        RE(ensure_connected(signal, mock=False, force_reconnect=True))
    assert (
        signal._connect_task
        and signal._connect_task.done()
        and signal._connect_task.exception()
    )


async def time_taken_by(coro) -> float:
    start = time.monotonic()
    await coro
    return time.monotonic() - start


async def test_set_and_wait_for_value_same_set_as_read():
    signal = epics_signal_rw(int, "pva://pv", name="signal")
    await signal.connect(mock=True)
    assert await signal.get_value() == 0
    set_mock_put_proceeds(signal, False)

    do_read_set = Event()
    callback_on_mock_put(signal, lambda *args, **kwargs: do_read_set.set())

    async def wait_and_set_proceeds():
        await do_read_set.wait()
        set_mock_put_proceeds(signal, True)

    async def check_set_and_wait():
        await (await set_and_wait_for_value(signal, 1, timeout=0.1))

    await asyncio.gather(wait_and_set_proceeds(), check_set_and_wait())
    assert await signal.get_value() == 1


async def test_set_and_wait_for_value_different_set_and_read():
    set_signal = epics_signal_rw(int, "pva://set", name="set-signal")
    read_signal = epics_signal_r(str, "pva://read", name="read-signal")
    await set_signal.connect(mock=True)
    await read_signal.connect(mock=True)

    do_read_set = Event()

    callback_on_mock_put(set_signal, lambda *args, **kwargs: do_read_set.set())

    async def wait_and_set_read():
        await do_read_set.wait()
        set_mock_value(read_signal, "test")

    async def check_set_and_wait():
        await (
            await set_and_wait_for_other_value(
                set_signal, 1, read_signal, "test", timeout=100
            )
        )

    await asyncio.gather(wait_and_set_read(), check_set_and_wait())
    assert await set_signal.get_value() == 1


async def test_set_and_wait_for_value_different_set_and_read_times_out():
    set_signal = epics_signal_rw(int, "pva://set", name="set-signal")
    read_signal = epics_signal_r(str, "pva://read", name="read-signal")
    await set_signal.connect(mock=True)
    await read_signal.connect(mock=True)

    do_read_set = Event()

    callback_on_mock_put(set_signal, lambda *args, **kwargs: do_read_set.set())

    async def wait_and_set_read():
        await do_read_set.wait()
        set_mock_value(read_signal, "not_test")

    async def check_set_and_wait():
        await (
            await set_and_wait_for_other_value(
                set_signal, 1, read_signal, "test", timeout=0.1
            )
        )

    with pytest.raises(TimeoutError):
        await asyncio.gather(wait_and_set_read(), check_set_and_wait())


async def test_wait_for_value_with_value():
    signal = epics_signal_rw(str, read_pv="pva://signal", name="signal")
    await signal.connect(mock=True)
    await signal.set("blah")

    with pytest.raises(
        asyncio.TimeoutError,
        match="signal didn't match 'something' in 0.1s, last value 'blah'",
    ):
        await wait_for_value(signal, "something", timeout=0.1)
    assert await time_taken_by(wait_for_value(signal, "blah", timeout=2)) < 0.1
    t = asyncio.create_task(
        time_taken_by(wait_for_value(signal, "something else", timeout=2))
    )
    await asyncio.sleep(0.2)
    assert not t.done()
    set_mock_value(signal, "something else")
    assert 0.2 < await t < 1.0


async def test_wait_for_value_with_funcion():
    signal = epics_signal_rw(float, read_pv="pva://signal", name="signal")
    await signal.connect(mock=True)
    set_mock_value(signal, 45.8)

    def less_than_42(v):
        return v < 42

    with pytest.raises(
        asyncio.TimeoutError,
        match="signal didn't match less_than_42 in 0.1s, last value 45.8",
    ):
        await wait_for_value(signal, less_than_42, timeout=0.1)
    t = asyncio.create_task(
        time_taken_by(wait_for_value(signal, less_than_42, timeout=2))
    )
    await asyncio.sleep(0.2)
    assert not t.done()
    set_mock_value(signal, 41)
    assert 0.2 < await t < 1.0
    assert await time_taken_by(wait_for_value(signal, less_than_42, timeout=2)) < 0.1


@pytest.mark.parametrize(
    "signal_method,signal_class",
    [(soft_signal_r_and_setter, SignalR), (soft_signal_rw, SignalRW)],
)
async def test_create_soft_signal(signal_method, signal_class):
    SIGNAL_NAME = "TEST-PREFIX:SIGNAL"
    INITIAL_VALUE = "INITIAL"
    if signal_method == soft_signal_r_and_setter:
        signal, _ = signal_method(str, INITIAL_VALUE, SIGNAL_NAME)
    elif signal_method == soft_signal_rw:
        signal = signal_method(str, INITIAL_VALUE, SIGNAL_NAME)
    else:
        raise ValueError(signal_method)
    assert signal.source == f"soft://{SIGNAL_NAME}"
    assert isinstance(signal, signal_class)
    await signal.connect()
    assert isinstance(signal._connector.backend, SoftSignalBackend)
    assert (await signal.get_value()) == INITIAL_VALUE


@pytest.fixture
async def mock_signal():
    mock_signal = epics_signal_rw(int, "pva://mock_signal", name="mock_signal")
    await mock_signal.connect(mock=True)
    yield mock_signal


async def test_assert_value(mock_signal: SignalRW):
    set_mock_value(mock_signal, 168)
    await assert_value(mock_signal, 168)


async def test_assert_reaading(mock_signal: SignalRW):
    set_mock_value(mock_signal, 888)
    dummy_reading = {
        "mock_signal": Reading({"alarm_severity": 0, "timestamp": ANY, "value": 888})
    }
    await assert_reading(mock_signal, dummy_reading)


async def test_failed_assert_reaading(mock_signal: SignalRW):
    set_mock_value(mock_signal, 888)
    dummy_reading = {
        "mock_signal": Reading({"alarm_severity": 0, "timestamp": ANY, "value": 88})
    }
    with pytest.raises(AssertionError):
        await assert_reading(mock_signal, dummy_reading)


class DummyReadable(StandardReadable):
    """A demo Readable to produce read and config signal"""

    def __init__(self, prefix: str, name="") -> None:
        # Define some signals
        with self.add_children_as_readables(Format.HINTED_SIGNAL):
            self.value = epics_signal_r(float, prefix + "Value")
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.mode = epics_signal_rw(str, prefix + "Mode")
            self.mode2 = epics_signal_rw(str, prefix + "Mode2")
        # Set name and signals for read() and read_configuration()
        super().__init__(name=name)


@pytest.fixture
async def mock_readable():
    async with DeviceCollector(mock=True):
        mock_readable = DummyReadable("SIM:READABLE:", name="mock_readable")

    yield mock_readable


async def test_assert_configuration(mock_readable: DummyReadable):
    set_mock_value(mock_readable.value, 123)
    set_mock_value(mock_readable.mode, "super mode")
    set_mock_value(mock_readable.mode2, "slow mode")
    dummy_config_reading = {
        "mock_readable-mode": (
            {
                "alarm_severity": 0,
                "timestamp": ANY,
                "value": "super mode",
            }
        ),
        "mock_readable-mode2": {
            "alarm_severity": 0,
            "timestamp": ANY,
            "value": "slow mode",
        },
    }
    await assert_configuration(mock_readable, dummy_config_reading)


async def test_signal_get_and_set_logging(caplog):
    caplog.set_level(logging.DEBUG)
    mock_signal_rw = epics_signal_rw(int, "pva://mock_signal", name="mock_signal")
    await mock_signal_rw.connect(mock=True)
    await mock_signal_rw.set(value=0)
    assert "Putting value 0 to backend at source" in caplog.text
    assert "Successfully put value 0 to backend at source" in caplog.text
    await mock_signal_rw.get_value()
    assert "get_value() on source" in caplog.text


async def test_subscription_logs(caplog):
    caplog.set_level(logging.DEBUG)
    mock_signal_rw = epics_signal_rw(int, "pva://mock_signal", name="mock_signal")
    await mock_signal_rw.connect(mock=True)
    cbs = []
    mock_signal_rw.subscribe(cbs.append)
    assert "Making subscription" in caplog.text
    mock_signal_rw.clear_sub(cbs.append)
    assert "Closing subscription on source" in caplog.text


async def test_signal_unknown_datatype():
    class SomeClass:
        def __init__(self):
            self.some_attribute = "some_attribute"

        def some_function(self):
            pass

    err_str = (
        "Can't make converter for <class "
        "'test_signal.test_signal_unknown_datatype.<locals>.SomeClass'>"
    )
    with pytest.raises(TypeError, match=err_str):
        await epics_signal_rw(SomeClass, "pva://mock_signal").connect(mock=True)
    with pytest.raises(TypeError, match=err_str):
        await epics_signal_rw(SomeClass, "ca://mock_signal").connect(mock=True)
    with pytest.raises(TypeError, match=err_str):
        soft_signal_rw(SomeClass)



from __future__ import annotations

import asyncio
import functools
from collections.abc import AsyncGenerator, Awaitable, Callable, Mapping
from typing import Any, Generic, cast
from unittest.mock import Mock

from bluesky.protocols import (
    Locatable,
    Location,
    Movable,
    Status,
    Subscribable,
)
from event_model import DataKey

from ._device import Device, DeviceConnector
from ._mock_signal_backend import MockSignalBackend
from ._protocol import (
    AsyncConfigurable,
    AsyncReadable,
    AsyncStageable,
    Reading,
)
from ._signal_backend import (
    SignalBackend,
    SignalDatatypeT,
    SignalDatatypeV,
)
from ._soft_signal_backend import SoftSignalBackend
from ._status import AsyncStatus
from ._utils import CALCULATE_TIMEOUT, DEFAULT_TIMEOUT, CalculatableTimeout, Callback, T

_mock_signal_backends: dict[Device, MockSignalBackend] = {}


async def _wait_for(coro: Awaitable[T], timeout: float | None, source: str) -> T:
    try:
        return await asyncio.wait_for(coro, timeout)
    except asyncio.TimeoutError as e:
        raise asyncio.TimeoutError(source) from e


def _add_timeout(func):
    @functools.wraps(func)
    async def wrapper(self: Signal, *args, **kwargs):
        return await _wait_for(func(self, *args, **kwargs), self._timeout, self.source)

    return wrapper


class SignalConnector(DeviceConnector):
    def __init__(self, backend: SignalBackend):
        self.backend = self._init_backend = backend

    async def connect(
        self,
        device: Device,
        mock: bool | Mock,
        timeout: float,
        force_reconnect: bool,
    ):
        if mock:
            self.backend = MockSignalBackend(self._init_backend, mock)
            _mock_signal_backends[device] = self.backend
        else:
            self.backend = self._init_backend
        device.log.debug(f"Connecting to {self.backend.source(device.name, read=True)}")
        await self.backend.connect(timeout)


class Signal(Device, Generic[SignalDatatypeT]):
    """A Device with the concept of a value, with R, RW, W and X flavours"""

    _connector: SignalConnector

    def __init__(
        self,
        backend: SignalBackend[SignalDatatypeT],
        timeout: float | None = DEFAULT_TIMEOUT,
        name: str = "",
    ) -> None:
        super().__init__(name=name, connector=SignalConnector(backend))
        self._timeout = timeout

    @property
    def source(self) -> str:
        """Like ca://PV_PREFIX:SIGNAL, or "" if not set"""
        return self._connector.backend.source(self.name, read=True)

    def __setattr__(self, name: str, value: Any) -> None:
        if name != "parent" and isinstance(value, Device):
            raise AttributeError(
                f"Cannot add Device or Signal {value} as a child of Signal {self}, "
                "make a subclass of Device instead"
            )
        return super().__setattr__(name, value)


class _SignalCache(Generic[SignalDatatypeT]):
    def __init__(self, backend: SignalBackend[SignalDatatypeT], signal: Signal):
        self._signal = signal
        self._staged = False
        self._listeners: dict[Callback, bool] = {}
        self._valid = asyncio.Event()
        self._reading: Reading[SignalDatatypeT] | None = None
        self.backend = backend
        signal.log.debug(f"Making subscription on source {signal.source}")
        backend.set_callback(self._callback)

    def close(self):
        self.backend.set_callback(None)
        self._signal.log.debug(f"Closing subscription on source {self._signal.source}")

    async def get_reading(self) -> Reading[SignalDatatypeT]:
        await self._valid.wait()
        assert self._reading is not None, "Monitor not working"
        return self._reading

    async def get_value(self) -> SignalDatatypeT:
        reading = await self.get_reading()
        return reading["value"]

    def _callback(self, reading: Reading[SignalDatatypeT]):
        self._signal.log.debug(
            f"Updated subscription: reading of source {self._signal.source} changed"
            f"from {self._reading} to {reading}"
        )
        self._reading = reading
        self._valid.set()
        for function, want_value in self._listeners.items():
            self._notify(function, want_value)

    def _notify(
        self,
        function: Callback[dict[str, Reading[SignalDatatypeT]] | SignalDatatypeT],
        want_value: bool,
    ):
        assert self._reading, "Monitor not working"
        if want_value:
            function(self._reading["value"])
        else:
            function({self._signal.name: self._reading})

    def subscribe(self, function: Callback, want_value: bool) -> None:
        self._listeners[function] = want_value
        if self._valid.is_set():
            self._notify(function, want_value)

    def unsubscribe(self, function: Callback) -> bool:
        self._listeners.pop(function)
        return self._staged or bool(self._listeners)

    def set_staged(self, staged: bool):
        self._staged = staged
        return self._staged or bool(self._listeners)


class SignalR(Signal[SignalDatatypeT], AsyncReadable, AsyncStageable, Subscribable):
    """Signal that can be read from and monitored"""

    _cache: _SignalCache | None = None

    def _backend_or_cache(
        self, cached: bool | None = None
    ) -> _SignalCache | SignalBackend:
        # If cached is None then calculate it based on whether we already have a cache
        if cached is None:
            cached = self._cache is not None
        if cached:
            assert self._cache, f"{self.source} not being monitored"
            return self._cache
        else:
            return self._connector.backend

    def _get_cache(self) -> _SignalCache:
        if not self._cache:
            self._cache = _SignalCache(self._connector.backend, self)
        return self._cache

    def _del_cache(self, needed: bool):
        if self._cache and not needed:
            self._cache.close()
            self._cache = None

    @_add_timeout
    async def read(self, cached: bool | None = None) -> dict[str, Reading]:
        """Return a single item dict with the reading in it"""
        return {self.name: await self._backend_or_cache(cached).get_reading()}

    @_add_timeout
    async def describe(self) -> dict[str, DataKey]:
        """Return a single item dict with the descriptor in it"""
        return {self.name: await self._connector.backend.get_datakey(self.source)}

    @_add_timeout
    async def get_value(self, cached: bool | None = None) -> SignalDatatypeT:
        """The current value"""
        value = await self._backend_or_cache(cached).get_value()
        self.log.debug(f"get_value() on source {self.source} returned {value}")
        return value

    def subscribe_value(self, function: Callback[SignalDatatypeT]):
        """Subscribe to updates in value of a device"""
        self._get_cache().subscribe(function, want_value=True)

    def subscribe(self, function: Callback[dict[str, Reading]]) -> None:
        """Subscribe to updates in the reading"""
        self._get_cache().subscribe(function, want_value=False)

    def clear_sub(self, function: Callback) -> None:
        """Remove a subscription."""
        self._del_cache(self._get_cache().unsubscribe(function))

    @AsyncStatus.wrap
    async def stage(self) -> None:
        """Start caching this signal"""
        self._get_cache().set_staged(True)

    @AsyncStatus.wrap
    async def unstage(self) -> None:
        """Stop caching this signal"""
        self._del_cache(self._get_cache().set_staged(False))


class SignalW(Signal[SignalDatatypeT], Movable):
    """Signal that can be set"""

    @AsyncStatus.wrap
    async def set(
        self,
        value: SignalDatatypeT,
        wait=True,
        timeout: CalculatableTimeout = CALCULATE_TIMEOUT,
    ) -> None:
        """Set the value and return a status saying when it's done"""
        if timeout == CALCULATE_TIMEOUT:
            timeout = self._timeout
        source = self._connector.backend.source(self.name, read=False)
        self.log.debug(f"Putting value {value} to backend at source {source}")
        await _wait_for(self._connector.backend.put(value, wait=wait), timeout, source)
        self.log.debug(f"Successfully put value {value} to backend at source {source}")


class SignalRW(SignalR[SignalDatatypeT], SignalW[SignalDatatypeT], Locatable):
    """Signal that can be both read and set"""

    @_add_timeout
    async def locate(self) -> Location:
        """Return the setpoint and readback."""
        setpoint, readback = await asyncio.gather(
            self._connector.backend.get_setpoint(), self._backend_or_cache().get_value()
        )
        return Location(setpoint=setpoint, readback=readback)


class SignalX(Signal):
    """Signal that puts the default value"""

    @AsyncStatus.wrap
    async def trigger(
        self, wait=True, timeout: CalculatableTimeout = CALCULATE_TIMEOUT
    ) -> None:
        """Trigger the action and return a status saying when it's done"""
        if timeout == CALCULATE_TIMEOUT:
            timeout = self._timeout
        source = self._connector.backend.source(self.name, read=False)
        self.log.debug(f"Putting default value to backend at source {source}")
        await _wait_for(self._connector.backend.put(None, wait=wait), timeout, source)
        self.log.debug(f"Successfully put default value to backend at source {source}")


def soft_signal_rw(
    datatype: type[SignalDatatypeT],
    initial_value: SignalDatatypeT | None = None,
    name: str = "",
    units: str | None = None,
    precision: int | None = None,
) -> SignalRW[SignalDatatypeT]:
    """Creates a read-writable Signal with a SoftSignalBackend.
    May pass metadata, which are propagated into describe.
    """
    backend = SoftSignalBackend(datatype, initial_value, units, precision)
    signal = SignalRW(backend=backend, name=name)
    return signal


def soft_signal_r_and_setter(
    datatype: type[SignalDatatypeT],
    initial_value: SignalDatatypeT | None = None,
    name: str = "",
    units: str | None = None,
    precision: int | None = None,
) -> tuple[SignalR[SignalDatatypeT], Callable[[SignalDatatypeT], None]]:
    """Returns a tuple of a read-only Signal and a callable through
    which the signal can be internally modified within the device.
    May pass metadata, which are propagated into describe.
    Use soft_signal_rw if you want a device that is externally modifiable
    """
    backend = SoftSignalBackend(datatype, initial_value, units, precision)
    signal = SignalR(backend=backend, name=name)
    return (signal, backend.set_value)


def _generate_assert_error_msg(name: str, expected_result, actual_result) -> str:
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    return (
        f"Expected {WARNING}{name}{ENDC} to produce"
        + f"\n{FAIL}{expected_result}{ENDC}"
        + f"\nbut actually got \n{FAIL}{actual_result}{ENDC}"
    )


async def assert_value(signal: SignalR[SignalDatatypeT], value: Any) -> None:
    """Assert a signal's value and compare it an expected signal.

    Parameters
    ----------
    signal:
        signal with get_value.
    value:
        The expected value from the signal.

    Notes
    -----
    Example usage::
        await assert_value(signal, value)

    """
    actual_value = await signal.get_value()
    assert actual_value == value, _generate_assert_error_msg(
        name=signal.name,
        expected_result=value,
        actual_result=actual_value,
    )


async def assert_reading(
    readable: AsyncReadable, expected_reading: Mapping[str, Reading]
) -> None:
    """Assert readings from readable.

    Parameters
    ----------
    readable:
        Callable with readable.read function that generate readings.

    reading:
        The expected readings from the readable.

    Notes
    -----
    Example usage::
        await assert_reading(readable, reading)

    """
    actual_reading = await readable.read()
    assert expected_reading == actual_reading, _generate_assert_error_msg(
        name=readable.name,
        expected_result=expected_reading,
        actual_result=actual_reading,
    )


async def assert_configuration(
    configurable: AsyncConfigurable,
    configuration: Mapping[str, Reading],
) -> None:
    """Assert readings from Configurable.

    Parameters
    ----------
    configurable:
        Configurable with Configurable.read function that generate readings.

    configuration:
        The expected readings from configurable.

    Notes
    -----
    Example usage::
        await assert_configuration(configurable configuration)

    """
    actual_configurable = await configurable.read_configuration()
    assert configuration == actual_configurable, _generate_assert_error_msg(
        name=configurable.name,
        expected_result=configuration,
        actual_result=actual_configurable,
    )


def assert_emitted(docs: Mapping[str, list[dict]], **numbers: int):
    """Assert emitted document generated by running a Bluesky plan

    Parameters
    ----------
    Doc:
        A dictionary

    numbers:
        expected emission in kwarg from

    Notes
    -----
    Example usage::
        assert_emitted(docs, start=1, descriptor=1,
        resource=1, datum=1, event=1, stop=1)
    """
    assert list(docs) == list(numbers), _generate_assert_error_msg(
        name="documents",
        expected_result=list(numbers),
        actual_result=list(docs),
    )
    actual_numbers = {name: len(d) for name, d in docs.items()}
    assert actual_numbers == numbers, _generate_assert_error_msg(
        name="emitted",
        expected_result=numbers,
        actual_result=actual_numbers,
    )


async def observe_value(
    signal: SignalR[SignalDatatypeT],
    timeout: float | None = None,
    done_status: Status | None = None,
) -> AsyncGenerator[SignalDatatypeT, None]:
    """Subscribe to the value of a signal so it can be iterated from.

    Parameters
    ----------
    signal:
        Call subscribe_value on this at the start, and clear_sub on it at the
        end
    timeout:
        If given, how long to wait for each updated value in seconds. If an update
        is not produced in this time then raise asyncio.TimeoutError
    done_status:
        If this status is complete, stop observing and make the iterator return.
        If it raises an exception then this exception will be raised by the iterator.

    Notes
    -----
    Example usage::

        async for value in observe_value(sig):
            do_something_with(value)
    """

    q: asyncio.Queue[SignalDatatypeT | Status] = asyncio.Queue()
    if timeout is None:
        get_value = q.get
    else:

        async def get_value():
            return await asyncio.wait_for(q.get(), timeout)

    if done_status is not None:
        done_status.add_callback(q.put_nowait)

    signal.subscribe_value(q.put_nowait)
    try:
        while True:
            item = await get_value()
            if done_status and item is done_status:
                if exc := done_status.exception():
                    raise exc
                else:
                    break
            else:
                yield cast(SignalDatatypeT, item)
    finally:
        signal.clear_sub(q.put_nowait)


class _ValueChecker(Generic[SignalDatatypeT]):
    def __init__(self, matcher: Callable[[SignalDatatypeT], bool], matcher_name: str):
        self._last_value: SignalDatatypeT | None = None
        self._matcher = matcher
        self._matcher_name = matcher_name

    async def _wait_for_value(self, signal: SignalR[SignalDatatypeT]):
        async for value in observe_value(signal):
            self._last_value = value
            if self._matcher(value):
                return

    async def wait_for_value(
        self, signal: SignalR[SignalDatatypeT], timeout: float | None
    ):
        try:
            await asyncio.wait_for(self._wait_for_value(signal), timeout)
        except asyncio.TimeoutError as e:
            raise asyncio.TimeoutError(
                f"{signal.name} didn't match {self._matcher_name} in {timeout}s, "
                f"last value {self._last_value!r}"
            ) from e


async def wait_for_value(
    signal: SignalR[SignalDatatypeT],
    match: SignalDatatypeT | Callable[[SignalDatatypeT], bool],
    timeout: float | None,
):
    """Wait for a signal to have a matching value.

    Parameters
    ----------
    signal:
        Call subscribe_value on this at the start, and clear_sub on it at the
        end
    match:
        If a callable, it should return True if the value matches. If not
        callable then value will be checked for equality with match.
    timeout:
        How long to wait for the value to match

    Notes
    -----
    Example usage::

        wait_for_value(device.acquiring, 1, timeout=1)

    Or::

        wait_for_value(device.num_captured, lambda v: v > 45, timeout=1)
    """
    if callable(match):
        checker = _ValueChecker(match, match.__name__)  # type: ignore
    else:
        checker = _ValueChecker(lambda v: v == match, repr(match))
    await checker.wait_for_value(signal, timeout)


async def set_and_wait_for_other_value(
    set_signal: SignalW[SignalDatatypeT],
    set_value: SignalDatatypeT,
    read_signal: SignalR[SignalDatatypeV],
    read_value: SignalDatatypeV,
    timeout: float = DEFAULT_TIMEOUT,
    set_timeout: float | None = None,
) -> AsyncStatus:
    """Set a signal and monitor another signal until it has the specified value.

    This function sets a set_signal to a specified set_value and waits for
    a read_signal to have the read_value.

    Parameters
    ----------
    signal:
        The signal to set
    set_value:
        The value to set it to
    read_signal:
        The signal to monitor
    read_value:
        The value to wait for
    timeout:
        How long to wait for the signal to have the value
    set_timeout:
        How long to wait for the set to complete

    Notes
    -----
    Example usage::

        set_and_wait_for_value(device.acquire, 1, device.acquire_rbv, 1)
    """
    # Start monitoring before the set to avoid a race condition
    values_gen = observe_value(read_signal)

    # Get the initial value from the monitor to make sure we've created it
    current_value = await anext(values_gen)

    status = set_signal.set(set_value, timeout=set_timeout)

    # If the value was the same as before no need to wait for it to change
    if current_value != read_value:

        async def _wait_for_value():
            async for value in values_gen:
                if value == read_value:
                    break

        try:
            await asyncio.wait_for(_wait_for_value(), timeout)
        except asyncio.TimeoutError as e:
            raise TimeoutError(
                f"{read_signal.name} didn't match {read_value} in {timeout}s"
            ) from e

    return status


async def set_and_wait_for_value(
    signal: SignalRW[SignalDatatypeT],
    value: SignalDatatypeT,
    timeout: float = DEFAULT_TIMEOUT,
    status_timeout: float | None = None,
) -> AsyncStatus:
    """Set a signal and monitor it until it has that value.

    Useful for busy record, or other Signals with pattern:
      - Set Signal with wait=True and stash the Status
      - Read the same Signal to check the operation has started
      - Return the Status so calling code can wait for operation to complete

    Parameters
    ----------
    signal:
        The signal to set
    value:
        The value to set it to
    timeout:
        How long to wait for the signal to have the value
    status_timeout:
        How long the returned Status will wait for the set to complete

    Notes
    -----
    Example usage::

        set_and_wait_for_value(device.acquire, 1)
    """
    return await set_and_wait_for_other_value(
        signal, value, signal, value, timeout, status_timeout
    )



from ._block import (
    BitMux,
    CommonPandaBlocks,
    DataBlock,
    PcapBlock,
    PcompBlock,
    PcompDirection,
    PulseBlock,
    SeqBlock,
    TimeUnits,
)
from ._control import PandaPcapController
from ._hdf_panda import HDFPanda
from ._table import (
    DatasetTable,
    PandaHdf5DatasetType,
    SeqTable,
    SeqTrigger,
)
from ._trigger import (
    PcompInfo,
    SeqTableInfo,
    StaticPcompTriggerLogic,
    StaticSeqTableTriggerLogic,
)
from ._utils import phase_sorter
from ._writer import PandaHDFWriter

__all__ = [
    "CommonPandaBlocks",
    "DataBlock",
    "BitMux",
    "PcapBlock",
    "PcompBlock",
    "PcompDirection",
    "PulseBlock",
    "SeqBlock",
    "TimeUnits",
    "HDFPanda",
    "PandaHDFWriter",
    "PandaPcapController",
    "DatasetTable",
    "PandaHdf5DatasetType",
    "SeqTable",
    "SeqTrigger",
    "PcompInfo",
    "SeqTableInfo",
    "StaticPcompTriggerLogic",
    "StaticSeqTableTriggerLogic",
    "phase_sorter",
]



from collections.abc import Sequence
from typing import Any, get_args, get_origin

import numpy as np

from ophyd_async.core import (
    SignalBackend,
    SignalDatatypeT,
    SubsetEnum,
    get_dtype,
    get_enum_cls,
)


def get_supported_values(
    pv: str,
    datatype: type,
    pv_choices: Sequence[str],
) -> dict[str, str]:
    enum_cls = get_enum_cls(datatype)
    if not enum_cls:
        raise TypeError(f"{datatype} is not an Enum")
    choices = [v.value for v in enum_cls]
    error_msg = f"{pv} has choices {pv_choices}, but {datatype} requested {choices} "
    if issubclass(enum_cls, SubsetEnum):
        if not set(choices).issubset(pv_choices):
            raise TypeError(error_msg + "to be a subset of them.")
    else:
        if set(choices) != set(pv_choices):
            raise TypeError(error_msg + "to be strictly equal to them.")

    # Take order from the pv choices
    supported_values = {x: x for x in pv_choices}
    # But override those that we specify via the datatype
    for v in enum_cls:
        supported_values[v.value] = v
    return supported_values


def format_datatype(datatype: Any) -> str:
    if get_origin(datatype) is np.ndarray and get_args(datatype)[0] == tuple[int]:
        dtype = get_dtype(datatype)
        return f"Array1D[np.{dtype.name}]"
    elif get_origin(datatype) is Sequence:
        return f"Sequence[{get_args(datatype)[0].__name__}]"
    elif isinstance(datatype, type):
        return datatype.__name__
    else:
        return str(datatype)


class EpicsSignalBackend(SignalBackend[SignalDatatypeT]):
    def __init__(
        self,
        datatype: type[SignalDatatypeT] | None,
        read_pv: str = "",
        write_pv: str = "",
    ):
        self.read_pv = read_pv
        self.write_pv = write_pv
        super().__init__(datatype)



from collections.abc import Sequence
from xml.etree import ElementTree as ET

import bluesky.plan_stubs as bps

from ophyd_async.core import Device
from ophyd_async.epics.adcore import (
    NDArrayBaseIO,
    NDAttributeDataType,
    NDAttributeParam,
    NDAttributePv,
    NDFileHDFIO,
)


def setup_ndattributes(
    device: NDArrayBaseIO, ndattributes: Sequence[NDAttributePv | NDAttributeParam]
):
    root = ET.Element("Attributes")

    for ndattribute in ndattributes:
        if isinstance(ndattribute, NDAttributeParam):
            ET.SubElement(
                root,
                "Attribute",
                name=ndattribute.name,
                type="PARAM",
                source=ndattribute.param,
                addr=str(ndattribute.addr),
                datatype=ndattribute.datatype.value,
                description=ndattribute.description,
            )
        elif isinstance(ndattribute, NDAttributePv):
            ET.SubElement(
                root,
                "Attribute",
                name=ndattribute.name,
                type="EPICS_PV",
                source=ndattribute.signal.source.split("ca://")[-1],
                dbrtype=ndattribute.dbrtype.value,
                description=ndattribute.description,
            )
        else:
            raise ValueError(
                f"Invalid type for ndattributes: {type(ndattribute)}. "
                "Expected NDAttributePv or NDAttributeParam."
            )
    xml_text = ET.tostring(root, encoding="unicode")
    yield from bps.abs_set(device.nd_attributes_file, xml_text, wait=True)


def setup_ndstats_sum(detector: Device):
    hdf = getattr(detector, "hdf", None)
    assert isinstance(hdf, NDFileHDFIO), (
        f"Expected {detector.name} to have 'hdf' attribute that is an NDFilHDFIO, "
        f"got {hdf}"
    )
    yield from (
        setup_ndattributes(
            hdf,
            [
                NDAttributeParam(
                    name=f"{detector.name}-sum",
                    param="NDPluginStatsTotal",
                    datatype=NDAttributeDataType.DOUBLE,
                    description="Sum of the array",
                )
            ],
        )
    )



from collections.abc import AsyncGenerator, AsyncIterator

from event_model import DataKey

from ophyd_async.core import DEFAULT_TIMEOUT, DetectorWriter, NameProvider, PathProvider

from ._pattern_generator import PatternGenerator


class PatternDetectorWriter(DetectorWriter):
    pattern_generator: PatternGenerator

    def __init__(
        self,
        pattern_generator: PatternGenerator,
        path_provider: PathProvider,
        name_provider: NameProvider,
    ) -> None:
        self.pattern_generator = pattern_generator
        self.path_provider = path_provider
        self.name_provider = name_provider

    async def open(self, multiplier: int = 1) -> dict[str, DataKey]:
        return await self.pattern_generator.open_file(
            self.path_provider, self.name_provider(), multiplier
        )

    async def close(self) -> None:
        self.pattern_generator.close()

    def collect_stream_docs(self, indices_written: int) -> AsyncIterator:
        return self.pattern_generator.collect_stream_docs(indices_written)

    async def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        async for index in self.pattern_generator.observe_indices_written(timeout):
            yield index

    async def get_indices_written(self) -> int:
        return self.pattern_generator.image_counter



import asyncio

from ophyd_async.core import (
    AsyncStatus,
    DetectorController,
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adcore

from ._kinetix_io import KinetixDriverIO, KinetixTriggerMode

KINETIX_TRIGGER_MODE_MAP = {
    DetectorTrigger.internal: KinetixTriggerMode.internal,
    DetectorTrigger.constant_gate: KinetixTriggerMode.gate,
    DetectorTrigger.variable_gate: KinetixTriggerMode.gate,
    DetectorTrigger.edge_trigger: KinetixTriggerMode.edge,
}


class KinetixController(DetectorController):
    def __init__(
        self,
        driver: KinetixDriverIO,
    ) -> None:
        self._drv = driver
        self._arm_status: AsyncStatus | None = None

    def get_deadtime(self, exposure: float | None) -> float:
        return 0.001

    async def prepare(self, trigger_info: TriggerInfo):
        await asyncio.gather(
            self._drv.trigger_mode.set(KINETIX_TRIGGER_MODE_MAP[trigger_info.trigger]),
            self._drv.num_images.set(trigger_info.total_number_of_triggers),
            self._drv.image_mode.set(adcore.ImageMode.multiple),
        )
        if trigger_info.livetime is not None and trigger_info.trigger not in [
            DetectorTrigger.variable_gate,
            DetectorTrigger.constant_gate,
        ]:
            await self._drv.acquire_time.set(trigger_info.livetime)

    async def arm(self):
        self._arm_status = await adcore.start_acquiring_driver_and_ensure_status(
            self._drv
        )

    async def wait_for_idle(self):
        if self._arm_status:
            await self._arm_status

    async def disarm(self):
        await adcore.stop_busy_record(self._drv.acquire, False, timeout=1)
